"""
Module for creating generators for accessing sequences of data from a DataFrame
"""

import logging
import random
import time
from typing import Any, List, Optional, Union

import keras.utils
import numpy as np
import pandas as pd
import polars as pl

from damast.core.types import DataFrame, XDataFrame

__all__ = [
    "GroupSequenceAccessor",
    "SequenceIterator"
]
logger = logging.getLogger("damast")


# https://www.tensorflow.org/tutorials/structured_data/time_series
class GroupSequenceAccessor:
    """
    A generator that allows access to a length-limited sequence of a particular group.

    The resulting dataset (X) will have a shape of:
    :code:`(<batch_size>, <sequence_length>, <number-of-features>)`.

    The resulting targets (y) if requires will lead to a label dataset of shape
    :code:`(<batch_size>, <sequence_length>, <number-of-targets>)`.

    Per default the dataset is assumed to be sorted, e.g., typically in time-based order.
    One can name however use ``sort_columns``, to require sorting so that a sequence becomes a valid timeline.
    From the overall group-based sequence a random subsequence of given length is sampled.

    :param df: The dataframe from which the data (train, test, ...) shall be extracted.

        .. note::
            This can be the combined dataframe for train, test, validate since the generator allows to limit the group
            ids from which will be selected

    :param group_column: the name of the column that identifies the group
    :param sort_columns: Names of the columns that shall be used for sorting - if not set, no sorting will be done
    :param timeout_in_s: Searching for a sequence of a given length might fail, since the dataset might not contain data
        of the given length.
    """
    DEFAULT_TIMEOUT_IN_S: float = 30.0

    def __init__(self,
                 df: DataFrame,
                 group_column: str,
                 sort_columns: List[str] = None,
                 timeout_in_s: int = DEFAULT_TIMEOUT_IN_S):
        self.df = df

        self.group_column = group_column
        self.groups = df.unique(group_column)

        if sort_columns is not None:
            self.sort_columns = sort_columns if type(sort_columns) == list else [sort_columns]
        else:
            self.sort_columns = sort_columns

        self.timeout_in_s = timeout_in_s

    def split_random(self, ratios: List[float]) -> List[List[Any]]:
        """
        Create ``N=len(ratios)`` groups of the dataframe, with given ratios, return the corresponding groups.

        The groups are based on :attr:`GroupSequenceAccessor.group_column`.

        :param ratios: List of relative partition sizes (will be normalized, so that all elements sum to 1
        :return: Following the ratios, returns lists of randomly sampled values from the id/group column
        """
        scaled_ratios = np.asarray(ratios) / sum(ratios)
        groups = self.groups[self.group_column].to_numpy().copy()

        random.shuffle(groups)
        number_of_groups = len(groups)

        partition_sizes = np.asarray(np.round(number_of_groups*scaled_ratios), dtype=int)
        assert (len(groups) == sum(partition_sizes))
        from_idx = 0
        partitions = []
        for ps in partition_sizes:
            to_idx = min(from_idx + ps, number_of_groups)
            partitions.append(groups[from_idx:to_idx])
            from_idx = to_idx

        return partitions

    def to_keras_generator(self, features: List[str],
                           target: List[str] = None,
                           groups: List[str] = None,
                           sequence_length: int = 50,
                           sequence_forecast: int = 0,
                           batch_size: int = 1024,
                           shuffle: bool = False,
                           infinite: bool = False,
                           verbose: bool = True) -> keras.utils.Sequence:
        """
        Create a batch generator suitable as a Keras datasource.

        By default, the generator is infinite, i.e. it loops continuously over the data.
        Thus, you need to specify the :code:`"steps_per_epoch"` arg when fitting a Keras model,
        the :code:`"validation_steps"` when using it for validation, and :code:`"steps"` when
        calling the :code:`"predict"` method of a keras model.

        :param features: A list of features.
        :param target: The dependent or target column or a list of columns, if any.
        :param groups: A list of group ids for which sequence generation will be done - this must be a subset
            of the existing group value in the dataframe, see :func:`split_random`
        :param sequence_length: Required length of the to-be-generated sequence
        :param sequence_forecast: If target is given, this is the length of a forecasted sequence -
            for a sequence-to-sequence generator
        :param batch_size: Number of samples per chunk of data. This can be thought of as the batch size.
        :param shuffle: If True, shuffle a sequence - if sort_columns is given, the setting will have no effect
        :param infinite: If True, the generator is infinite, i.e. it loops continuously over the data.
            If False, the generator does only one pass over the data.
        :param verbose: If True, show an info on the recommended :code:`"steps_per_epoch"`
            based on the total number of samples and :code:`"batch_size"`.

        Example:

            .. highlight:: python
            .. code-block:: python

                from damast.data_handling.accessors import GroupSequenceAccessor
                import tensorflow.keras as K

                df = ...
                features = ['lat', 'lon']
                target = ['nav_status']

                gsa = GroupSequenceAccessor(df=df, group_column="mmsi", sort_columns=["timestamp"])
                train_ids, validate_ids, test_ids = gsa.split_random(ratios=[0.8, 0.1, 0.1])

                # Create a training generator
                train_generator = gsa.to_keras_generator(features=features, target=target,
                                                        sequence_length=50, sequence_forecast=1, batch_size=10)

                # Build a recurrent neural network model to deal with the sequence, e.g.,
                # to forecast the next sequence element
                nn_model = K.Sequential()
                nn_model.add(K.layers.SimpleRNN(2, return_sequences=True, input_shape=[50, 2])
                nn_model.add(K.layers.SimpleRNN(2, return_sequences=True))
                nn_model.add(K.layers.SimpleRNN(2))
                nn_model.compile(optimizer='sgd', loss='mse')

                nn_model.fit(x=train_generator, epochs=3, steps_per_epoch=645)
        """
        if verbose:
            current_level = logger.getEffectiveLevel()
            logger.setLevel(logging.INFO)
            steps_per_epoch = np.ceil(len(self.df) / batch_size)
            logger.info(f'Recommended {steps_per_epoch=}')
            logger.setLevel(current_level)

        # Sanity checks before creating generator
        # Check that all features have the same data-type
        datatypes = [XDataFrame(self.df).dtype(f) for f in features]
        for dtype in datatypes:
            if dtype != datatypes[0]:
                raise ValueError(f"{self.__class__.__name__}:"
                                 f" Features {features} do not have a consistent (single) datatype,"
                                 f" got {datatypes}")

        use_target = target is not None
        if use_target:
            datatypes = [XDataFrame(self.df).dtype(t) for t in target]
            for dtype in datatypes:
                if dtype != datatypes[0]:
                    raise ValueError(f"{self.__class__.__name__}:"
                                     f" Targets {target} do not have a consistent (single) datatype,"
                                     f" got {datatypes}")

        if use_target:
            target = target if type(target) == list else [target]

            if sequence_forecast < 0:
                raise ValueError(f"{self.__class__.__name__}: Sequence forecast cannot be negative")
            if sequence_forecast == 0:
                raise ValueError(f"{self.__class__.__name__}: Cannot do extract targets with no sequence forecast")
        else:
            if sequence_forecast > 0:
                raise ValueError(f"{self.__class__.__name__}: Cannot do sequence forecast with no targets")

        if self.sort_columns is not None and shuffle:
            raise RuntimeError(f"{self.__class__.__name__}: Cannot sort and shuffle sequence at the same time")

        def _generator(features: List[str], target: Optional[List[str]],
                       groups: List[Any],
                       sequence_length: int, sequence_forecast: int,
                       chunk_size: int, shuffle: bool, infinite: bool,
                       ):
            """
            A generator function to yield the next sequence.

            If `targets` is not supplied, the generator has unlabeled data.
            Otherwise, the generator will provide labels for forecasting, i.e.
            taking samples from the group (after sequence length).

            :param features: List of feature name
            :param target: List of targets
            :param groups: List of groups from which sequences can be generated
            :param sequence_length: Required length of a sequence
            :param sequence_forecast: Length of the target sequence
            :param chunk_size: Number of sequences in a chunk (aka batch)
            :param shuffle: Whether the data (rows) in a sequence should be shuffled
            :param infinite: Run the generator infinitely, i.e. requires functions that use this generator to define
                a stopping criteria, e.g., steps_in_epoch
            """
            # Gather all columns in one list
            all_columns = features.copy()
            if self.sort_columns is not None:
                all_columns += self.sort_columns
            use_target = target is not None
            if use_target:
                target = target if type(target) == list else target
                all_columns += target
            else:
                # If no target, we are not forecasting
                sequence_forecast = 0
            all_columns = list(set(all_columns))

            if groups is None:
                groups = self.groups

            while True:
                chunk = []
                target_chunk = []
                for i in range(chunk_size):
                    sequence: DataFrame = None
                    start_time = time.perf_counter()
                    sample_count = 0
                    sample_length = 0

                    # Find a valid subsequence, i.e. one with the given length
                    while not (time.perf_counter() - start_time) > self.timeout_in_s:
                        group = random.choice(groups)
                        # Since we will need the timeline later - we further deal with pandas DataFrame
                        # directly - thus, we do not use a copy of the DataFrame (only used columns)
                        sequence = self.df\
                                    .filter(pl.col(self.group_column) == group[self.group_column])\
                                    .select(all_columns)

                        # If sort columns are set, then ensure that the sorting is done
                        if self.sort_columns is not None:
                            sequence = sequence.sort(by=self.sort_columns)
                        elif shuffle:
                            sequence = sequence.sample(fraction=1)

                        if isinstance(sequence, DataFrame):
                            sequence = sequence.collect()

                        len_sequence = sequence.shape[0]
                        if len_sequence >= (sequence_length + sequence_forecast):
                            break

                        sample_count += 1
                        sample_length += len_sequence
                    if sequence is None or len_sequence < (sequence_length + sequence_forecast):
                        raise RuntimeError(f"{self.__class__.__name__}: could not identify a sufficiently long sequence"
                                           f" within given timeout of {self.timeout_in_s}:"
                                           f" mean length was {sample_length / sample_count}")

                    max_start_idx = len(sequence) - (sequence_length + sequence_forecast)
                    if max_start_idx == 0:
                        start_idx = 0
                    else:
                        start_idx = random.randint(0, max_start_idx)

                    sequence_window = sequence[start_idx:start_idx + sequence_length]
                    chunk.append(sequence_window[features].to_numpy())
                    # Target is a list of output "labels" in the ML sense
                    if use_target:
                        # Target is taken from the message "after" those in the window.
                        target_start_idx = start_idx + sequence_length
                        target_end_idx = target_start_idx + sequence_forecast

                        target_window = sequence[target_start_idx:target_end_idx][target]
                        if sequence_forecast == 1:
                            target_window = target_window[0]

                        # target it the last step in the timeline, so the last
                        target_chunk.append(target_window.to_numpy())

                X = np.array(chunk)
                if use_target:
                    if np.lib.NumpyVersion(np.__version__) >= '2.0.0':
                        y = np.array(target_chunk)
                    else:
                        y = np.array(target_chunk, copy=False)
                    yield (X, y)
                else:
                    yield (X,)
                if not infinite:
                    break

        return _generator(features=features, target=target,
                          groups=groups, sequence_length=sequence_length, sequence_forecast=sequence_forecast,
                          chunk_size=batch_size, shuffle=shuffle, infinite=infinite)


class SequenceIterator:
    """
    A generator that allows iterate over the windows of a length-limited sequence.

    The resulting dataset (X) will have a shape of:
        (<sequence_length>, <number-of-features>).

    The resulting targets (y) if requires will lead to a label dataset of shape
        (<sequence_length>, <number-of-targets>)

    Per default the dataset is assumed to be sorted, e.g., typically in time-based order.
    One can name however use :attr:`"sort_columns"`, to require sorting so that a sequence becomes a valid timeline.

    :param df: The dataframe from which the data (train, test, ...) shall be extracted.
    :param sort_columns: Names of the columns that shall be used for sorting - if None, no sorting will be done
    """

    df: Union[DataFrame, pd.DataFrame]

    def __init__(self,
                 df: Union[DataFrame, pd.DataFrame],
                 sort_columns: List[str] = None):
        if sort_columns is not None:
            if isinstance(df, DataFrame):
                df = df.collect()

            if isinstance(df, pl.dataframe.DataFrame):
                self.df = df.sort(by=sort_columns)
            elif isinstance(df, pd.DataFrame):
                self.df = df.sort_values(by=sort_columns)
            else:
                raise RuntimeError(f"{self.__class__.__name__}.__init__: df was {type(df)}, but must be "
                                   f"either polars.dataframe.DataFrame or pandas.DataFrame")
        else:
            self.df = df

    def to_keras_generator(self, features,
                           target=None,
                           sequence_length: int = 50,
                           sequence_forecast: int = 1) -> keras.utils.Sequence:
        """
        Create a batch generator suitable as a Keras datasource.

        By default, the generator is infinite, i.e. it loops continuously over the data.
        Thus, you need to specify the :attr:`"steps_per_epoch"` arg when fitting a Keras model,
        the :attr:`"validation_steps"` when using it for validation, and :attr:`"steps"` when
        calling the :attr:`"predict"` method of a keras model.

        :param features: A list of features.
        :param target: The dependent or target column or a list of columns, if any.
        :param sequence_length: Length of input sequence
        :param sequence_forecast: If target is given, this is the length of a forecasted sequence -
            for a sequence-to-sequence generator

        Example:

            .. highlight:: python
            .. code-block:: python

                from damast.data_handling.accessors import GroupSequenceAccessor
                import tensorflow.keras as K

                df = ...
                features = ['lat', 'lon']
                target = ['nav_status']

                it = SequenceIterator(df=df, sort_columns=["timestamp"])
                # Create a training generator
                it = it.to_keras_generator(features=features, target=target, sequence_length=50, sequence_forecast=1)

                # Build a recurrent neural network model to deal with the sequence, e.g.,
                # to forecast the next sequence element
                nn_model = K.Sequential()
                nn_model.add(K.layers.SimpleRNN(2, return_sequences=True, input_shape=[50, 2])
                nn_model.add(K.layers.SimpleRNN(2, return_sequences=True))
                nn_model.add(K.layers.SimpleRNN(2))
                nn_model.compile(optimizer='sgd', loss='mse')

                nn_model.fit(x=train_generator, epochs=3, steps_per_epoch=645)
        """

        # Sanity checks before creating generator
        # Check that all features have the same data-type
        datatypes = [XDataFrame(self.df).dtype(f) for f in features]
        for dtype in datatypes:
            if dtype != datatypes[0]:
                raise ValueError(f"{self.__class__.__name__}:"
                                 f" Features {features} do not have a consistent (single) datatype,"
                                 f" got {datatypes}")

        use_target = target is not None
        if use_target:
            datatypes = [XDataFrame(self.df).dtype(t) for t in target]
            for dtype in datatypes:
                if dtype != datatypes[0]:
                    raise ValueError(f"{self.__class__.__name__}:"
                                     f" Targets {target} do not have a consistent (single) datatype,"
                                     f" got {datatypes}")

        if sequence_forecast < 0:
            raise ValueError(f"{self.__class__.__name__}: Sequence forecast cannot be negative")

        if use_target:
            target = target if type(target) == list else target
            if sequence_forecast == 0:
                raise ValueError(f"{self.__class__.__name__}: Cannot do extract targets with no sequence forecast")
        else:
            if sequence_forecast > 0:
                raise ValueError(f"{self.__class__.__name__}: Cannot do sequence forecast with no targets")

        sequence = self.df

        # Since we will need the timeline later - we further deal with
        # the actual data
        if not isinstance(sequence, pl.dataframe.DataFrame):
            sequence = sequence.collect()

        len_sequence = len(sequence)  # Equivalent to sequence.shape[0]
        if len_sequence < (sequence_length + sequence_forecast):
            raise RuntimeError(f"{self.__class__.__name__}:"
                               f" 'Sequence length' plus 'forecast length' ({sequence_length + sequence_forecast})"
                               f" larger than dataframe of size ({len_sequence})")

        def _generator(sequence: pl.dataframe.DataFrame,
                       features: List[str],
                       target: Optional[List[str]],
                       sequence_length: int,
                       sequence_forecast: int):
            """
            A generator function to yield the next sequence.

            If ``sequence_forecast>=1`` then this generator returns a tuple ``(X, y)`` where
            `X` is a dataframe of shape `(sequence_length, len(features)`,
            `y` is a dataframe of shape `(sequence_forecast, len(target)`
            selected randomly from

            :param sequence: The time-series
            :param features: List of feature name
            :param target: List of targets (labels)
            :param sequence_length: Required length of a sequence
            :param sequence_forecast: Length of the target sequence
            :param chunk_size: Number of sequences in a chunk (aka batch)
            :param shuffle: Whether the data (rows) in a sequence should be shuffled
            :param infinite: Run the generator infinitely, i.e. requires functions that use this generator to define
                             a stopping criteria, e.g., steps_in_epoch

            """
            max_start_idx = len_sequence - (sequence_length + sequence_forecast) + 1
            use_target = target is not None

            # (Num_batches=1, Length of Sequences, Number of features)
            X = np.empty((sequence_length, len(features)), dtype=XDataFrame(sequence).dtype(features[0]).to_python())
            if target is not None:
                y = np.empty((sequence_forecast, len(target)), dtype=XDataFrame(sequence).dtype(target[0]).to_python())

            # Iterate through the windowed sequences until the index is exhausted
            for start_idx in range(max_start_idx):

                # Extract features
                X[:, :] = sequence[start_idx:start_idx + sequence_length][features].to_numpy()

                if use_target:
                    target_start_idx = start_idx + sequence_length
                    target_end_idx = target_start_idx + sequence_forecast
                    # Extract targets
                    y[:, :] = sequence[target_start_idx:target_end_idx][target].to_numpy()

                if use_target:
                    yield (X, y)
                else:
                    yield (X,)

        return _generator(sequence=sequence, features=features, target=target,
                          sequence_length=sequence_length, sequence_forecast=sequence_forecast,
                          )



"""
This module extends the existing units found in ::class::`astropy.units`.
"""
import astropy.units as _units

__all__ = [
    "knots",
    "units",
    "unit_registry",
    "Unit"
]

units = _units
Unit = _units.Unit
# Additional type

# nautical mile in [nmi] = [NM]
nmi = units.def_unit("nmi", 1.852 * units.km)
knots = units.def_unit("knots", nmi / units.hour)

units.add_enabled_units([nmi, knots])

unit_registry = units.get_current_unit_registry().registry



import re
from argparse import Action, ArgumentParser
from pathlib import Path

from damast.cli.base import BaseParser
from damast.core.annotations import Annotation
from damast.core.dataframe import DAMAST_SPEC_SUFFIX, AnnotatedDataFrame
from damast.core.metadata import DataSpecification, MetaData


class SetTxtFieldAction(Action):
    def __call__(self, parser, namespace, values, option_string=None):
        name = option_string.lstrip('--set-')
        if not hasattr(namespace, "update_metadata"):
            setattr(namespace, "update_metadata", MetaData(columns=[]))

        for value in values:
            column, column_value = value.split(":", 1)
            if column not in namespace.update_metadata:
                ds = DataSpecification(name=column)
                namespace.update_metadata.columns.append(ds)

            setattr(namespace.update_metadata[column], name, column_value)

class DataAnnotateParser(BaseParser):
    """
    Argparser for creation of the specification / annotating an AnnotatedDataFrame

    :param parser: The base parser
    """

    def __init__(self, parser: ArgumentParser):
        super().__init__(parser=parser)

        parser.description = "damast annotate - extract (default) or apply annotation to dataset"
        parser.add_argument("-f", "--files",
                            help="Files or patterns of the (annotated) data file that should be annotated",
                            nargs="+",
                            type=str,
                            required=True
                            )
        parser.add_argument("-o", "--output-dir",
                            help="Output directory",
                            default=None,
                            required=False)

        parser.add_argument("--output-spec-file",
                            help="The spec file name - if provided with path, it will override output-dir",
                            default=None,
                            required=False)

        str_options = ["description", "abbreviation", "unit"]
        for field_name in str_options:
            parser.add_argument(f"--set-{field_name}",
                                nargs="+",
                                action=SetTxtFieldAction,
                                metavar="COLUMN:VALUE",
                                help=f"Set {field_name} in spec for a column and value",
                                required=False)

        parser.add_argument("--inplace",
                            help="Update the dataset inplace (only possible for a single file)",
                            action="store_true",
                            required=False)
        parser.add_argument("--apply",
                help="Update the annotation inference an rewrite the metadata to the dataset",
                action="store_true",
                required=False,
        )



    def execute(self, args):
        super().execute(args)

        files_stats = self.get_files_stats(args.files)
        print(f"Loading dataframe ({files_stats.number_of_files} files) of total size: {files_stats.total_size} MB")

        adf = AnnotatedDataFrame.from_files(files=args.files, metadata_required=False)
        print(adf.head(10).collect())

        metadata_filename = MetaData.specfile(args.files)
        output_dir = Path(metadata_filename).parent
        if args.output_dir is not None:
            output_dir = Path(args.output_dir)
            output_dir.mkdir(parents=True, exist_ok=True)

        if args.output_spec_file:
            spec_file = Path(args.output_spec_file.replace(DAMAST_SPEC_SUFFIX, ''))
            if spec_file.is_absolute():
                spec_file.parent.mkdir(parents=True, exist_ok=True)
                metadata_filename = spec_file.with_suffix(DAMAST_SPEC_SUFFIX)
            else:
                metadata_filename = output_dir / spec_file.with_suffix(DAMAST_SPEC_SUFFIX)
        else:
            metadata_filename = output_dir / metadata_filename.name

        metadata = AnnotatedDataFrame.infer_annotation(df=adf)

        if hasattr(args, "update_metadata"):
            metadata = metadata.merge(args.update_metadata)

        metadata.add_annotation(
                Annotation(
                    name=Annotation.Key.Source,
                    value=args.files
                )
        )
        metadata.save_yaml(metadata_filename)

        if hasattr(args, "update_metadata") or args.apply:
            if len(args.files) == 1:
                output_file = args.files[0]
                adf._metadata = metadata
                if not args.inplace:
                    output_file = output_dir / Path(output_file).name
                    print(f"Creating {output_file}")
                else:
                    print(f"Updating {output_file}")
                adf.export(output_file)
            else:
                raise ValueError("Cannot update metadata for multiple files")

        print(f"Created {metadata_filename}")



"""
Data Processing Module

Contains AIS-specific pipelines for filtering and augmenting data
"""


from pathlib import Path
from typing import Any, Dict, List, Union

from damast.core import DataProcessingPipeline
from damast.core.types import XDataFrame
from damast.data_handling.exploration import plot_lat_lon
from damast.data_handling.transformers import (
    AddTimestamp,
    AddUndefinedValue,
    ChangeTypeColumn,
    DropMissingOrNan,
    FilterWithin,
    RemoveValueRows,
    )
from damast.data_handling.transformers.augmenters import AddLocalIndex
from damast.data_handling.transformers.visualisers import PlotLatLon
from damast.domains.maritime.ais import vessel_types
from damast.domains.maritime.transformers import AddVesselType, ComputeClosestAnchorage

from .data_specification import ColumnName

__all__ = ["CleanseAndSanitise", "DataProcessing"]


ParamsType = Dict[str, Any]


def get_outputs_dir(workdir: Union[str, Path]) -> Path:
    return Path(workdir) / "processed_data"


def get_plots_dir(workdir: Union[str, Path]) -> Path:
    return Path(workdir) / "plots"


class CleanseAndSanitise(DataProcessingPipeline):
    """
    A :class:`damast.core.DataProcessingPipeline` for cleaning and sanitising AIS data.

    :param message_types: List of allowable message types (all other message types will be removed)
    :param columns_default_values: Dictionary mapping column names to default values. All columns
        in this dict will have all empty rows replaced with this value
    :param columns_compress_type: Dictionary mapping column names to compressed data-types. All columns
        in this dict will have a new column with the compressed data
    :param workdir: Base directory towards which transformer output which be relative
    :param name: Name of pipeline.

    The expected input columns are:

    * :attr:`damast.domains.maritime.data_specification.ColumnName.MESSAGE_TYPE`
    * :attr:`damast.domains.maritime.data_specification.ColumnName.DATE_TIME_UTC`
    * :attr:`damast.domains.maritime.data_specification.ColumnName.SPEED_OVER_GROUND`
    * :attr:`damast.domains.maritime.data_specification.ColumnName.COURSE_OVER_GROUND`
    * Any column in ``columns_default_values`` and in ``columns_compress_types``.

    The pipeline performs the following actions

    1. Keep only data from satellite, i.e. :code:`source != 'g'`
    2. Remove row which do not have a timestamp
    3. Keep only message-types in input ``message_types``
    4. Create a :code:`"timestamp"` field from date time UTC
    5. Replace NaN/Na with default values as defined in  :code:`"columns_default_values"`
    6. Create new columns with new types using :code:`columns_compress_types`.
       The new column is name from the original column name with "_newtype" as a suffix.

    """

    def __init__(self, message_types: List[int],
                 columns_default_values: Dict[str, Any],
                 columns_compress_types: Dict[str, str],
                 workdir: Union[str, Path],
                 name: str = "Cleanse and sanitise data",
                 name_mappings: Dict[str, str] = {}):
        super().__init__(name=name,
                         base_dir=workdir,
                         name_mappings=name_mappings)

        self.add("Remove rows with ground as source", RemoveValueRows("g"),
                 name_mappings={"x": ColumnName.SOURCE})
        self.add("Remove rows with null dates", DropMissingOrNan(),
                 name_mappings={"x": ColumnName.DATE_TIME_UTC})
        self.add("Filter rows within message types", FilterWithin(message_types),
                 name_mappings={"x": ColumnName.MESSAGE_TYPE})
        self.add("Add Timestamp column to each row", AddTimestamp(),
                 name_mappings={"from": ColumnName.DATE_TIME_UTC, "to": ColumnName.TIMESTAMP})

        for column in columns_default_values:
            default_value = columns_default_values[column]
            self.add(f"Set default values to {column}", AddUndefinedValue(default_value),
                     name_mappings={"x": column})

        for column in columns_compress_types:
            new_type = columns_compress_types[column]
            self.add(f"Compress column {column}", ChangeTypeColumn(new_type),
                     name_mappings={"x": column, "y": f"{column}_{new_type}"})


class DataProcessing(DataProcessingPipeline):
    """
    Create an AIS-specific pipeline to an annotated dataframe.

    :param vessel_type_hdf5: Path to HDF5 file containing vessel types
    :param fishing_vessel_type_hdf5: Path to HDF5 file containing fishing vessel types
    :param anchorages_hdf5: Path to HDF5 file containing anchorages' coordinates
    :param workdir: Base directory towards which transformer output which be relative
    :param name: Name of pipeline.

    .. note::
        Initializing this class plots the input of ``anchorages_hdf5`` which should probably
        be moved elsewhere

    The pipeline requires the following input columns:

    * :attr:`damast.domains.maritime.data_specification.ColumnName.MMSI`
    * :attr:`damast.domains.maritime.data_specification.ColumnName.LATITUDE`
    * :attr:`damast.domains.maritime.data_specification.ColumnName.LONGITUDE`
    * :attr:`damast.domains.maritime.data_specification.ColumnName.MMSI`
    * :attr:`damast.domains.maritime.data_specification.ColumnName.TIMESTAMP`

    The pipeline does the following actions:

    * Plots longitude and latitude of all data
    * Adds vessel-type information from ``vessel_type_hdf5`` to column
      :attr:`damast.domains.maritime.data_specification.ColumnName.VESSEL_TYPE`.
    * Replace missing vessel-types with integer representation of
      :class:`damast.domains.maritime.ais.vessel_types.Unspecified`
    * Add global fishing-watch vessel type from ``fishing_vessel_type_hdf5`` to column
      :attr:`damast.domains.maritime.data_specification.ColumnName.FISHING_TYPE`
    * Replace missing values in gfw vessel type with `-1`
    * Compute closest distance to anchorages from ``anchorages_hdf5`` and att to column
      :attr:`damast.domains.maritime.data_specification.ColumnName.DISTANCE_CLOSEST_ANCHORAGE`
    * Group messages by :attr:`damast.domains.maritime.data_specification.ColumnName.MMSI` and
      local group index based on :attr:`damast.domains.maritime.data_specification.ColumnName.TIMESTAMP`.
      Column name for group index and reverse index is
      :attr:`damast.domains.maritime.data_specification.ColumnName.HISTORIC_SIZE` and
      :attr:`damast.domains.maritime.data_specification.ColumnName.HISTORIC_SIZE_REVERSE`, respectively.

    """

    def __init__(self, workdir: Union[str, Path],
                 vessel_type_hdf5: Union[str, Path],
                 fishing_vessel_type_hdf5: Union[str, Path],
                 anchorages_hdf5: Union[str, Path],
                 name: str = "AIS-processor",
                 name_mappings: Dict[str, str] = {}):

        super().__init__(name=name,
                         base_dir=workdir,
                         name_mappings=name_mappings)

        plots_dir = get_plots_dir(workdir=workdir)
        plots_dir.mkdir(parents=True, exist_ok=True)

        # Output an overview over the existing anchorages
        anchorages_data = XDataFrame.open(anchorages_hdf5)

        plot_lat_lon(df=anchorages_data,
                     latitude_name="latitude",
                     longitude_name="longitude",
                     output_dir=plots_dir,
                     filename_prefix="anchorages-lat-lon")

        # Temporary converters to csv to be compatible
        if isinstance(vessel_type_hdf5, Path):
            vessel_type_csv = XDataFrame.open(vessel_type_hdf5)
        else:
            vessel_type_csv = vessel_type_hdf5

        if isinstance(anchorages_hdf5, Path):
            anchorages_csv = XDataFrame.open(anchorages_hdf5)
        else:
            anchorages_csv = anchorages_hdf5
        if isinstance(fishing_vessel_type_hdf5, Path):
            fishing_vessel_type_csv = XDataFrame.open(fishing_vessel_type_hdf5)
        else:
            fishing_vessel_type_csv = fishing_vessel_type_hdf5

        # FIXME: Name mapping does not apply here
        # self.add("plot_input-lat_lon",
        #         PlotLatLon(output_dir=plots_dir,
        #                    filename_prefix="lat-lon-input"),
        #         name_mappings={"LAT": ColumnName.LATITUDE,
        #                        "LON": ColumnName.LONGITUDE})

        self.add("augment_vessel_type",
                 AddVesselType(dataset=vessel_type_csv,
                               right_on=ColumnName.MMSI,
                               dataset_column=ColumnName.VESSEL_TYPE),
                 name_mappings={"x": ColumnName.MMSI,
                                "out": ColumnName.VESSEL_TYPE})
        self.add("Replace missing", AddUndefinedValue(vessel_types.Unspecified.to_id()),
                 name_mappings={"x": ColumnName.VESSEL_TYPE})

        self.add("augment_fishing_vessel_type",
                 AddVesselType(dataset=fishing_vessel_type_csv,
                               right_on=ColumnName.MMSI.lower(),
                               dataset_column=ColumnName.VESSEL_TYPE_GFW),
                 name_mappings={"x": ColumnName.MMSI,
                                "out": ColumnName.FISHING_TYPE})
        self.add("Replace missing", AddUndefinedValue(-1),
                 name_mappings={"x": ColumnName.FISHING_TYPE})
        self.add("augment_distance_to_closest_anchorage",
                 ComputeClosestAnchorage(dataset=anchorages_csv,
                                         columns=["latitude", "longitude"]),
                 name_mappings={"x": ColumnName.LATITUDE,
                                "y": ColumnName.LONGITUDE,
                                "distance": ColumnName.DISTANCE_CLOSEST_ANCHORAGE})
        self.add("Compute local message index",  AddLocalIndex(),
                 name_mappings={"group": ColumnName.MMSI,
                                "sort": ColumnName.TIMESTAMP,
                                "local_index": ColumnName.HISTORIC_SIZE,
                                "reverse_{{local_index}}": ColumnName.HISTORIC_SIZE_REVERSE})



import datetime as dt
import logging
import re
from argparse import ArgumentParser
from pathlib import Path

import polars as pl

from damast.cli.base import BaseParser
from damast.core.dataframe import AnnotatedDataFrame
from damast.utils.io import Archive

logger = logging.getLogger(__name__)

class DataInspectParser(BaseParser):
    """
    Argparser for inspecting AnnotatedDataFrame

    :param parser: The base parser
    """

    def __init__(self, parser: ArgumentParser):
        super().__init__(parser=parser)

        parser.description = "damast inspect - data inspection subcommand called"
        parser.add_argument("-f", "--files",
                            help="Files or patterns of the (annotated) data file that should be inspected (space separated)",
                            nargs="+",
                            type=str,
                            required=True
                            )

        parser.add_argument("--filter",
                help="Filter based on column data, e.g., mmsi==120123",
                action="append"
        )
        parser.add_argument("--head", type=int, default=10, help="First this number of rows, default is 10")
        parser.add_argument("--tail", type=int, default=10, help="Print number of rows from the end, default is 10")
        parser.add_argument("--column-count", type=int, default=10, help="Number of columns to show")

        parser.add_argument("--columns",
                help="Show/Select these columns",
                nargs="+",
                type=str,
                required=False
        )

    def expand_filter_arg(self, adf: AnnotatedDataFrame, arg: str):
        if arg in adf.column_names:
            return f"pl.col('{arg}')"

        m = re.match(r"datetime\((.*)\)", arg)
        if m:
            return f"dt.datetime.fromisoformat({m.group(1)})"

        return arg

    def execute(self, args):
        super().execute(args)

        files_stats = self.get_files_stats(args.files)
        print(f"Loading dataframe ({files_stats.number_of_files} files) of total size: {files_stats.total_size} MB")

        try:
            with Archive(filenames=args.files) as input_files:
                files = [x for x in input_files if AnnotatedDataFrame.get_supported_format(Path(x).suffix)]
                if not files:
                    raise RuntimeError(f"Inspection is not supported for input files: {input_files=}")

                adf = AnnotatedDataFrame.from_files(files=files, metadata_required=False)

                if args.filter:
                    filter_values = ""
                    for filter_expression in args.filter:
                        m = re.match(r"([^=<>]+)([=><]+)([^=<>]*)", filter_expression)
                        if m:
                            lhs = m.group(1).strip()
                            op = m.group(2).strip()
                            rhs = m.group(3).strip()

                            lhs = self.expand_filter_arg(adf, lhs)
                            rhs = self.expand_filter_arg(adf, rhs)

                            print(f"   .filter({lhs} {op} {rhs})")
                            filter_values += f".filter({lhs} {op} {rhs})"
                        else:
                            logger.warning(f"Filter expression invalid: {filter_expression}")
                    adf._dataframe = eval(f"adf._dataframe{filter_values}")
                    adf._metadata = AnnotatedDataFrame.infer_annotation(adf._dataframe)

                print(adf.metadata.to_str(columns=args.columns))
                print(f"\n\nFirst {args.head} and last {args.tail} rows:")
                df = adf._dataframe
                if args.columns:
                    df = df.select(args.columns)

                with pl.Config(tbl_rows=args.head, tbl_cols=args.column_count):
                    print(df.head(n=args.head).collect())
                with pl.Config(tbl_rows=args.tail, tbl_cols=args.column_count):
                    print(df.tail(n=args.tail).collect())

        except RuntimeError as e:
            if re.search(r"metadata is missing", str(e)) is not None:
                print(e)
            else:
                raise



"""
Module which collect all filters that filter the existing data.
"""

from typing import Any

import polars as pl
import polars.selectors as cs

import damast.core
from damast.core import AnnotatedDataFrame
from damast.core.dataprocessing import PipelineElement
from damast.core.types import XDataFrame

__all__ = [
    "FilterWithin",
    "RemoveValueRows"
]


class RemoveValueRows(PipelineElement):
    """
    Remove rows that do not have a defined value for a given column.

    :param remove_value: remove rows with this value.
    """
    _remove_value: Any

    def __init__(self, remove_value: Any):
        super().__init__()
        self._remove_value = remove_value

    @property
    def remove_value(self):
        return self._remove_value

    @damast.core.describe("Remove rows where a column has a specific value")
    @damast.core.input({"x": {}})
    @damast.core.output({"x": {}})
    def transform(self, df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        """
        Delete rows with remove_values
        """
        mapped_name = self.get_name("x")

        df._dataframe = df._dataframe.filter(
            pl.col(mapped_name) != self._remove_value
        )
        return df


class DropMissingOrNan(PipelineElement):
    """
    Drop rows that do not have a defined value or NaN for a given column.
    """

    @damast.core.describe("Drop rows where this column has a missing (or nan) value")
    @damast.core.input({"x": {}})
    @damast.core.output({"x": {}})
    def transform(self, df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        """
        Drop rows with missing value
        """
        mapped_name = self.get_name("x")
        dataframe = df._dataframe

        new_dataframe = dataframe.drop_nulls(subset=mapped_name)
        dtype = XDataFrame(new_dataframe).dtype(mapped_name)
        if dtype not in [str, pl.String]:
            new_dataframe = new_dataframe.drop_nans(subset=mapped_name)

        df._dataframe = new_dataframe
        return df


class FilterWithin(PipelineElement):
    """
    Filter rows and keep those within given values.

    :param within_values: list of values to keep
    """
    _within_values: Any

    def __init__(self, within_values: Any):
        super().__init__()
        self._within_values = within_values

    @property
    def within_values(self):
        return self._within_values

    @damast.core.describe("Filter rows and keep those within given values")
    @damast.core.input({"x": {}})
    @damast.core.output({"x": {}})
    def transform(self, df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        """
        Filter rows and keep those within given values
        """
        mapped_name = self.get_name("x")
        dataframe = df._dataframe
        new_dataframe = dataframe.filter(
            pl.col(mapped_name).is_in(self._within_values)
        )
        df._dataframe = new_dataframe
        return df



"""
Module which collects transformers that add / augment the existing data
"""
import datetime
import logging
from enum import Enum
from logging import Logger, getLogger
from pathlib import Path
from typing import Any, Union

import numpy as np
import numpy.typing as npt
import pandas as pd
import polars as pl
from sklearn.neighbors import BallTree

import damast.core
from damast.core import AnnotatedDataFrame
from damast.core.dataprocessing import PipelineElement
from damast.core.types import DataFrame, XDataFrame

__all__ = [
    "AddDeltaTime",
    "AddLocalIndex",
    "AddTimestamp",
    "AddUndefinedValue",
    "BallTreeAugmenter",
    "JoinDataFrameByColumn",
    "MultiplyValue"
]

_log: Logger = getLogger(__name__)


class JoinDataFrameByColumn(PipelineElement):
    """
    Add a column to an input dataframe by merging it with another dataset.

    :param dataset: Path to `.csv/.hdf5`-file or a `polars.dataframe.LazyFrame`.
    :param right_on: Column from `dataset` to use for joining data
    :param dataset_column: Name of column in `dataset` to add
    :param col_name: Name of augmented column
    :param sep: Separator in CSV file

    .. note::
        :code:`right_on` will not be added as a new column in the transformed dataset
    """

    class JoinHowType(str, Enum):
        INNER = 'inner'
        LEFT = 'left'
        RIGHT = 'right'
        FULL = 'full'
        SEMI = 'semi'
        ANTI = 'anti'
        CROSS = 'cross'

    right_on: str
    dataset_column: str
    join_how: JoinHowType

    dataset: DataFrame | Path

    def __init__(self,
                 dataset: Union[str, Path, XDataFrame],
                 right_on: str,
                 dataset_column: str,
                 how: JoinHowType = JoinHowType.LEFT,
                 sep: str = ";"):
        self.right_on: str = right_on
        self.dataset_column = dataset_column
        self.join_how = how

        # Load vessel type map
        if type(dataset) in [str, Path]:
            dataset = self.load_data(filename=dataset, sep=sep)

        # Check that the columns exist in dataset
        for col_name in [self.right_on, self.dataset_column]:
            if col_name not in dataset.columns:
                raise KeyError(f"Missing column: '{col_name}' in vessel type information: '{dataset.head()}'"
                               " - available are {','.join(vessel_type_data.columns)}")

        column_dtype = XDataFrame(dataset).dtype(self.dataset_column)
        if int not in [column_dtype, column_dtype.to_python()]:
            raise ValueError(f"{self.__class__.__name__}.__init__:"
                             f" column '{self.dataset_column}' must be of type int, "
                             f", but was: {column_dtype}")

        self.dataset = dataset

    @classmethod
    def load_data(cls,
                  filename: Union[str, Path], sep: str) -> DataFrame:
        """
        Load dataset from file

        :param filename: The input file (or path)
        :param sep: Separator in csv
        :return: A `DataFrame` with the data
        """
        try:
            return XDataFrame.open(path=filename, sep=sep)
        except FileNotFoundError as e:
            raise RuntimeError(f"{cls}: Vessel type information not accessible -- {e}")

    @damast.core.input({"x": {}})
    @damast.core.output({"out": {}})
    def transform(self, df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        """
        Join datasets by column "x".
        Adds column "out".

        :returns: DataFrame with added column
        """
        other_df = self.dataset.select(
                pl.col(self.right_on),
                pl.col(self.dataset_column)
        ).lazy()

        dataframe = df._dataframe.join(
                other=other_df,
                left_on=self.get_name("x"),
                right_on=self.right_on,
                #validate='1:1',
                #https://docs.pola.rs/api/python/stable/reference/dataframe/api/polars.DataFrame.join.html
                how=self.join_how.value
            )

        df._dataframe = dataframe.rename({self.dataset_column: self.get_name("out")})
        return df


class BallTreeAugmenter():
    """
    A class for distance computation using BallTree.

    Uses the `sklearn.neighbours.BallTree` to compute the distance for any n-dimensional
    feature. The BallTree is created prior to being passed in as the lambda function of a
    `DataFrame.add_virtual_column`. The object can later be depickled from the state,
    and one can retrieve any meta-data added to the class after construction.

    :param x: The points to use in the BallTree
    :param metric: The metric to use in the BallTree, for available metrics see:
        https://docs.scipy.org/doc/scipy/reference/spatial.distance.html
    """

    _tree: BallTree
    _metric: str
    _modified: datetime.datetime

    def __init__(self, x: npt.NDArray[np.float64], metric: str):
        self._tree = BallTree(x, metric=metric)
        self._metric = metric
        self.__name__ = f"Balltree_{self._metric}"
        self._modified = datetime.datetime.now()

    def update_balltree(self, x: npt.NDArray[np.float64]):
        """
        Replace points in the Balltree

        :param x: (npt.NDArray[np.float64]): The new points
        """
        logger = logging.getLogger("damast")
        if not np.allclose(self._tree.get_arrays()[0], x):
            logger.debug("Recreating balltree with new inputs")
            self._tree = BallTree(x, metric=self._metric)
            self._modified = datetime.datetime.now()
        else:
            logger.debug("No points to update in balltree")

    def __call__(self, x: npt.NDArray[np.float64], y: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
        """
        Compute distances between the Balltree and each entry in `x`"""
        return self._tree.query(np.vstack([x, y]).T, return_distance=True)[
            0].reshape(-1)

    @property
    def modified(self) -> datetime.datetime:
        """
        Last time the underlying BallTree was modified
        """
        return self._modified


class AddUndefinedValue(PipelineElement):
    """
    Replace missing and Not Available (NA) entries in a column with a given value.

    :param fill_value: The value replacing NA
    """
    _fill_value: Any

    def __init__(self, fill_value: Any):
        """Constructor"""
        self._fill_value = fill_value

    @property
    def fill_value(self):
        return self._fill_value

    @damast.core.describe("Fill undefined values")
    @damast.core.input({"x": {}})
    @damast.core.output({"x": {}})
    def transform(self, df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        """
        Fill in values for NA and missing entries
        """
        mapped_name = self.get_name("x")
        df._dataframe = df._dataframe.with_columns(
                pl.col(mapped_name).fill_null(self.fill_value).alias(mapped_name),
            ).with_columns(
                pl.col(mapped_name).fill_nan(self.fill_value).alias(mapped_name)
            )
        return df


class AddLocalIndex(PipelineElement):
    """
    Compute the local index of an entry in a given group (sorted by a given column).

    Also compute the reverse index, i.e. how many entries in the group are after this message
    """

    @damast.core.describe("Compute the local index for a timeseries (group)")
    @damast.core.input({"group": {"representation_type": int},
                        "sort": {}})
    @damast.core.output({"local_index": {"representation_type": int},
                         "reverse_{{local_index}}": {"representation_type": int}})
    def transform(self, df: damast.core.AnnotatedDataFrame) -> damast.core.AnnotatedDataFrame:
        dataframe = df._dataframe

        group_column = self.get_name("group")
        sort_column = self.get_name("sort")

        df._dataframe = dataframe\
                .sort(group_column, sort_column)\
                .with_columns(
                    pl.int_range(pl.len()).over(group_column).alias(self.get_name("local_index")),
                    pl.int_range(start=pl.len()-1, end=-1, step=-1).over(group_column).alias(self.get_name("reverse_{{local_index}}"))
                )

        return df


class AddDeltaTime(PipelineElement):
    """
    Compute the delta time between two subsequent timestamps
    """

    @damast.core.describe("Compute the delta time in s between two subsequent timestamps")
    @damast.core.input({"group": {"representation_type": int},
                        "time_column": {}})
    @damast.core.output({"delta_time": {"representation_type": float}})
    def transform(self, df: damast.core.AnnotatedDataFrame) -> damast.core.AnnotatedDataFrame:
        dataframe = df._dataframe

        group_column = self.get_name("group")
        time_column = self.get_name("time_column")

        dataframe = dataframe\
               .sort(group_column, time_column)\
               .with_columns(
                    pl.col(time_column).shift(1).alias('previous_timestamp')
               )

        dataframe = dataframe\
             .sort(group_column, time_column)\
             .with_columns(
                 (pl.col(time_column) - pl.col('previous_timestamp')).alias('delta_time')
             ).drop("previous_timestamp")

        dataframe = dataframe\
             .sort(group_column, time_column)\
             .with_columns(
                pl.col('delta_time').fill_null(pl.duration(seconds=0)).alias('delta_time')
             )

        df._dataframe = dataframe
        return df


def convert_to_datetime(date_string: str) -> float:
    """
    Convert date-time to timestamp (float)

    :param date_string: String representation of date, expected format is
        ``YYYY-MM-DD HH:MM:SS``
    :return: Time-stamp as float. Returns `nan`
    """
    try:
        return datetime.datetime.timestamp(
            datetime.datetime.strptime(
                date_string, "%Y-%m-%d %H:%M:%S"))
    except TypeError:
        return float(np.nan)


class AddTimestamp(PipelineElement):
    """
    Add Timestamp from date Time UTC.

    If time-stamp is not supplied for a row add ``NaN``
    """

    @damast.core.describe("Add Timestamp")
    @damast.core.input({"from": {"representation_type": str}})
    @damast.core.output({"to": {"representation_type": float}})
    def transform(self, df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        """
        Add Timestamp from datetimeUTC
        """
        from_mapped_name = self.get_name("from")
        to_mapped_name = self.get_name("to")

        df._dataframe = df._dataframe.with_columns(
                pl.col(from_mapped_name).map_elements(convert_to_datetime, return_dtype=float).alias(to_mapped_name)
        )
        return df


class MultiplyValue(PipelineElement):
    """
    Multiply a column by a value.

    :param multiply_value: The value to use to multiply
    """
    _mul_value: Any

    def __init__(self, mul_value: Any):
        """Constructor"""
        self._mul_value = mul_value

    @property
    def mul_value(self):
        return self._mul_value

    @damast.core.describe("Multiply a column by a given values")
    @damast.core.input({"x": {}})
    @damast.core.output({"x": {}})
    def transform(self, df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        """
        Multiply a column by a given value
        """
        mapped_name = self.get_name("x")
        df._dataframe = df._dataframe.with_columns(
            (pl.col(mapped_name)*self.mul_value).alias(mapped_name)
        )
        return df


class ChangeTypeColumn(PipelineElement):
    """
    Create a new column with the new type of a given column.

    The new column name can be defined by providing a name_mapping for a column 'y'.
    If no name_mapping is provided the column's new name will be 'y'

    :param new_type: The new type of the column
    """
    new_type: Any

    def __init__(self, new_type: Any):
        """Constructor"""
        if isinstance(new_type, pl.datatypes.classes.DataTypeClass):
            self.new_type = new_type
        elif type(new_type) == str:
            polar_type = new_type.capitalize()
            if not hasattr(pl.datatypes.classes, polar_type):
                raise TypeError(f"Type {new_type} has not correspondence in 'polars'")

            self.new_type = getattr(pl.datatypes.classes, polar_type)

    @damast.core.describe("Create a new column from an existing column but with a new type")
    @damast.core.input({"x": {}})
    @damast.core.output({"y": {}})
    def transform(self, df: AnnotatedDataFrame) -> AnnotatedDataFrame:
        """
        Change the default type of a column
        """
        input_mapped_name = self.get_name("x")
        output_mapped_name = self.get_name("y")

        df._dataframe = df._dataframe.with_columns(
            pl.col(input_mapped_name).cast(self.new_type).alias(output_mapped_name)
        )
        return df



from __future__ import annotations

import gc
import glob
import importlib
from abc import ABC, abstractmethod
from pathlib import Path
from tempfile import gettempdir
from typing import ClassVar, NamedTuple, OrderedDict

import keras
import keras.callbacks
import keras.utils
import pandas as pd

from damast.core import DataSpecification
from damast.core.types import DataFrame, XDataFrame

__all__ = [
    "BaseModel",
    "ModelInstanceDescription"
]

HISTORY_FILENAME = 'training-history.csv'
CHECKPOINT_BEST = "checkpoint.best.weights.h5"
MODEL_TF_HDF5 = "model.tf.hdf5"


class BaseModel(ABC):
    """
    BaseModel for Machine Learning Models

    :param name: A name for this model - by default this will be the class name
    :param features: List of the input feature names
    :param targets: List of the target feature names
    :param distribution_strategy: A tensorflow distribution strategy when trying to distribute the training across
                         multiple devices
    :param output_dir: model / experiment specific output directory
    """

    #: Input specification for this model
    input_specs: ClassVar[OrderedDict[str, DataSpecification]]
    features: list[str]

    #: Output specification for this model
    output_specs: ClassVar[OrderedDict[str, DataSpecification]]
    targets: list[str]

    #: Name of this model
    name: str
    #: Underlying keras model
    model: keras.Model

    # Computation
    distribution_strategy: object | None

    #: The directory where all the model output will be stored
    model_dir: Path

    #: History of the training
    history: keras.callbacks.History

    def __init__(self,
                 features: list[str],
                 targets: list[str] | None = None,
                 name: str | None = None,
                 distribution_strategy = None,
                 output_dir: Path = Path(gettempdir())
                 ):
        if not hasattr(self, "input_specs"):
            raise RuntimeError(f"{self.__class__.__name__}.__init__: input_specs are not set")

        if not hasattr(self, "output_specs"):
            raise RuntimeError(f"{self.__class__.__name__}.__init__: output_specs are not set")

        if name is None:
            self.name = self.__class__.__name__

        if distribution_strategy is None:
            if keras.backend.backend() == "tensorflow":
                import tensorflow as tf
                self.distribution_strategy = tf.distribute.get_strategy()
            else:
                self.distribution_strategy = None

        if name is None:
            name = self.__class__.__name__

        if features is not None:
            self.features = features

        if targets is not None:
            self.targets = targets

        self.name = name
        self.model_dir = output_dir / self.name
        self.model_dir.mkdir(parents=True, exist_ok=True)

        self._init_model()
        assert self.model is not None

    @abstractmethod
    def _init_model(self):
        """
        Initialize the internal keras model, i.e. self.model.
        """
        pass

    def load_weights(self,
                     checkpoint_filepath: str | Path):
        """
        Load the model weight from an existing checkpoint named by the checkpoint_filepath

        :param checkpoint_filepath: Name of the checkpoint filepath (prefix)
        """
        if self.model is None:
            raise RuntimeError(
                f"{self.__class__}.load_weights: you need to initialize the model first via _init_model()")

        self.model.load_weights(checkpoint_filepath)

    def build(self,
              optimizer: str | keras.optimizers.Optimizer = keras.optimizers.Adam(0.001),
              loss_function: str | keras.losses.Loss = "mse",
              metrics: list[str] | list[keras.metrics.Metric] = ["mse"],
              **kwargs):
        """
        Build / Compile the actual model

        :param optimizer: Name of the optimizer or an instance of keras.optimizers.Optimizer
        :param loss_function:  Name of the loss function or an instance of keras.losses.Loss
        :param metrics: List of names of the metrics that shall be used, or list of instance of keras.metrics.Metrics
        :param kwargs: additional arguments that can be forwarded to keras.engine.training.Model.compile
        """
        def _build_impl():
            # Plotting of the model
            self.plot()

            # https://keras.io/api/models/model_training_apis/
            self.model.compile(
                optimizer=optimizer,
                loss=loss_function,
                # see
                metrics=metrics,
                **kwargs
            )
            self.model.summary()

        if self.distribution_strategy and keras.backend.backend() == "tensorflow":
            with self.distribution_strategy.scope():
                _build_impl()
        else:
            _build_impl()

    def plot(self,
             suffix=".png") -> Path:
        """
        Plot the current model instance.

        :param suffix: Suffix that shall be used for the plot file
        :return: Path to the plotted file
        """

        plots_outputdir = self.model_dir / "plots"
        plots_outputdir.mkdir(parents=True, exist_ok=True)

        filename = plots_outputdir / f"{self.name}{suffix}"
        keras.utils.plot_model(model=self.model,
                               to_file=str(filename),
                               show_shapes=True,
                               expand_nested=True)
        return filename

    @property
    def checkpoints_dir(self) -> Path:
        """
        Directory for the checkpoints.

        :return: Path to the checkpoints directory
        """
        return self.model_dir / "checkpoints"

    @property
    def evaluation_file(self) -> Path:
        """
        Path to the evaluation file storing the results

        :return: evaluation file
        """
        return self.model_dir / f'evaluation-{CHECKPOINT_BEST}.csv'

    def train(self,
              training_data: tf.data.Dataset,
              validation_data: tf.data.Dataset,
              monitor: str = "val_loss",
              mode: str = "min",
              epochs: int = 1,
              initial_epoch: int = 0,
              save_history: bool = True,
              **kwargs) -> None:
        try:
            callbacks = []
            # check https://keras.io/api/callbacks/ for available callbacks
            checkpoint_filepath = self.checkpoints_dir / CHECKPOINT_BEST
            if keras.backend.backend() == "tensorflow":
                # https://www.tensorflow.org/tensorboard/get_started
                tensorboard_cb = keras.callbacks.TensorBoard(log_dir=str(self.model_dir / "logs"),
                                                             histogram_freq=1
                                                             )
                callbacks.append(tensorboard_cb)

            model_checkpoint_cb = keras.callbacks.ModelCheckpoint(
                filepath=checkpoint_filepath,
                save_weights_only=True,
                monitor=monitor,
                mode=mode,
                save_best_only=True)
            callbacks.append(model_checkpoint_cb)

            early_stopping_cb = keras.callbacks.EarlyStopping(patience=10,
                                                              restore_best_weights=True)
            callbacks.append(early_stopping_cb)

            # log_hyperparams_cb = hp.KerasCallback(logdir=self.model_dir / "logs",
            #                                       hparams)

            class CleanupCallback(keras.callbacks.Callback):
                def on_train_batch_begin(self, epoch, logs=None):
                    gc.collect()
            cleanup_cb = CleanupCallback()
            callbacks.append(cleanup_cb)

            # Allow to checkpoint per epoch
            self.history = self.model.fit(training_data,
                                          validation_data=validation_data,
                                          epochs=epochs,
                                          initial_epoch=initial_epoch,
                                          callbacks=callbacks,
                                          **kwargs
                                          )
        except Exception as e:
            raise RuntimeError(f"{self.__class__.__name__}: Training of {self.name} failed") from e

        # Save the history
        if save_history:
            history: DataFrame = XDataFrame.from_dict(self.history.history)
            history.write_csv(f"{self.model_dir / HISTORY_FILENAME}")

    def get_evaluations(self) -> DataFrame:
        """
        Read the evaluation results from the evaluation file.

        :raise FileNotFoundError: If evaluation result file does not exist
        """
        if self.evaluation_file.exists():
            return XDataFrame.read_csv(self.evaluation_file)

        raise FileNotFoundError(f"There is are no evaluation results for {self.name} available: {self.evaluation_file} "
                                "does not exist")

    def save(self) -> Path:
        """
        Save the model in the model directory.

        :return: Path to the saved model.
        """
        filename = self.model_dir / MODEL_TF_HDF5
        self.model.save(filepath=filename)
        return filename

    def evaluate(self,
                 label: str,
                 evaluation_data: tf.data.Dataset,
                 **kwargs) -> dict[str, any]:
        """
        Evaluate this model.

        :param label: Custom label for this evaluation, e.g., can be the name of the dataset being used
        :param evaluation_data: The data that shall be used for evaluation
        :return: DataFrame listing all performed evaluation results for this model
        """
        # Load the best weight
        checkpoint_best = self.checkpoints_dir / CHECKPOINT_BEST
        if glob.glob(f"{checkpoint_best}*"):
            self.load_weights(checkpoint_filepath=checkpoint_best)
        else:
            raise RuntimeError(f"{self.__class__}.evaluate: could not find the checkpoint data"
                               f" with best weights: {checkpoint_best} does not exist")

        # Evaluate the model against the provided data using the weights/checkpoint
        # with the (so far) best performance
        evaluation: dict[str, float] = self.model.evaluate(evaluation_data,
                                                           return_dict=True,
                                                           verbose=0,
                                                           **kwargs)

        evaluation_column_names = ["dataset"]
        evaluation_results: list[any] = [label]
        for name, value in evaluation.items():
            evaluation_column_names.append(name)
            evaluation_results.append(value)

        result_df = pd.DataFrame([evaluation_results], columns=evaluation_column_names)
        if self.evaluation_file.exists():
            result_df.to_csv(self.evaluation_file)

        return evaluation


class ModelInstanceDescription(NamedTuple):
    """
    Provide a description of a model instance to allow serialization.
    """
    model: BaseModel
    parameters: dict[str, str]

    @classmethod
    def from_dict(cls, data: dict[str, any]) -> ModelInstanceDescription:
        if "module_name" not in data:
            raise ValueError(f"{cls.__name__}.from_dict: missing 'module' specification")
        if "class_name" not in data:
            raise ValueError(f"{cls.__name__}.from_dict: missing 'class' specification")

        module_name = data["module_name"]
        class_name = data["class_name"]

        m_module = importlib.import_module(module_name)
        if not hasattr(m_module, class_name):
            raise ImportError(f"{cls.__name__}.from_dict: could not import '{class_name}' from '{m_module}'")

        model = getattr(m_module, class_name)

        parameters = data.get("parameters", {})
        return ModelInstanceDescription(model=model,
                                        parameters=parameters)

    def __iter__(self):
        yield "module_name", self.model.__module__
        yield "class_name", self.model.__qualname__
        yield "parameters", self.parameters



"""
Module containing (de)normalisation functions
"""
import numpy as np
import numpy.typing as npt
from numba import njit

__all__ = [
    "cyclic_denormalisation",
    "cyclic_normalisation",
    "normalize"
]


@njit
def normalize(
        x: npt.NDArray[np.float64], x_min: float, x_max: float, a: float, b: float
) -> npt.NDArray[np.float64]:
    """
    Normalize data in array `x` with lower bound `x_min` and upper bound `x_max`
    to be in the range `[a, b]`

    .. math::

        x_n = (b-a)\\frac{x-x_{min}}{x_{max}-x_{min}} + a

    :param x: Input array
    :param x_min: Minimum bound of input data
    :param x_max: Maximum bound of input data
    :param a: Minimum bound of output data
    :param b: Maximum bound of output data

    :returns: Normalized data
    """
    return (b - a) * (x - x_min) / (x_max - x_min) + a


def cyclic_normalisation(x: npt.NDArray[np.float64],
                         x_min: float,
                         x_max: float) -> npt.NDArray[np.float64]:
    """ Returns the data cycli-normalised between 0 and 1 """
    return np.array(
        [
            np.sin(2 * np.pi * normalize(x, x_min, x_max, 0, 1)),
            np.cos(2 * np.pi * normalize(x, x_min, x_max, 0, 1))
        ]
    )


def cyclic_denormalisation(x_sin: npt.NDArray[np.float64],
                           x_cos: npt.NDArray[np.float64],
                           min: float,
                           max: float) -> npt.NDArray[np.float64]:
    """ Returns the data cycli-denormalised"""
    # Flatten the input
    original_shape = x_sin.shape
    x_sin = x_sin.flatten()
    x_cos = x_cos.flatten()

    # Limit the min and max
    x_sin = np.clip(x_sin, -1, 1)
    x_cos = np.clip(x_cos, -1, 1)

    # Convert input between 0, 2PI -> -PI, PI
    x_sin = np.sin(np.arcsin(x_sin) - np.pi)
    x_cos = np.cos(np.arccos(x_cos) - np.pi)

    # Denormalization
    x_hat_normed = np.where(x_sin > 0, np.arccos(x_cos), - np.arccos(x_cos))
    hat = normalize(normalize(x_hat_normed, -np.pi, np.pi, 0, 1), 0, 1, min, max)

    # Reshape the output as the input
    return hat.reshape(original_shape)



"""
This namespace contains the core modules that are required to define metadata.
"""

from .annotations import Annotation
from .data_description import DataRange, MinMax
from .dataframe import AnnotatedDataFrame
from .dataprocessing import (
    DECORATED_ARTIFACT_SPECS,
    DECORATED_DESCRIPTION,
    DECORATED_INPUT_SPECS,
    DECORATED_OUTPUT_SPECS,
    DataProcessingPipeline,
    artifacts,
    describe,
    input,
    output,
    )
from .metadata import (
    ArtifactSpecification,
    DataSpecification,
    History,
    MetaData,
    ValidationMode,
    )

__all__ = [
    "AnnotatedDataFrame",
    "Annotation",
    "ArtifactSpecification",
    "DECORATED_ARTIFACT_SPECS",
    "DECORATED_DESCRIPTION",
    "DECORATED_INPUT_SPECS",
    "DECORATED_OUTPUT_SPECS",
    "DataProcessingPipeline",
    "DataRange",
    "DataSpecification",
    "History",
    "MetaData",
    "MinMax",
    "ValidationMode",
    "artifacts",
    "describe",
    "input",
    "output",
]



import warnings

import numpy as np
import numpy.typing as npt

__all__ = [
    "N_sigma",
    "N_sigma_limited"
]


def N_sigma_limited(x: npt.NDArray[np.float64],
                    N: int,
                    p: float = 0.99) -> npt.NDArray[np.float64]:
    """ Compute and return the N_sigma given a list without considering the
    :math:`100\\cdot(1-p)\\%` highest values

    :param x: The input array
    :param N: The number of standard deviations to add to the median
    :param p: The threshold (0 to 1) of values to include from the sorted array
    :returns: The median of the truncated input plus `N` times the standard deviation

    """
    warnings.warn("It is uncertain if this function gives you any sensible quantity")
    limit = x.quantile(p)
    return x.loc[x < limit].median() + N * x.loc[x < limit].std()


def N_sigma(x: npt.NDArray[np.float64], N: int) -> npt.NDArray[np.float64]:
    """ Compute and return the N_sigma given a list
        :param x: The input array

    :param N: The number of standard deviations to add to the median
    :returns: The median of the input `x` plus `N` times the standard deviation
    """
    warnings.warn("It is uncertain if this function gives you any sensible quantity")
    return x.median() + N * x.std()



"""
Collection of generic Transformer implementations
"""
from .augmenters import (
    AddDeltaTime,
    AddTimestamp,
    AddUndefinedValue,
    BallTreeAugmenter,
    ChangeTypeColumn,
    JoinDataFrameByColumn,
    MultiplyValue,
    )
from .filters import DropMissingOrNan, FilterWithin, RemoveValueRows
from .normalizers import normalize

__all__ = [
    "AddDeltaTime",
    "AddTimestamp",
    "AddUndefinedValue",
    "BallTreeAugmenter",
    "ChangeTypeColumn",
    "DropMissingOrNan",
    "FilterWithin",
    "JoinDataFrameByColumn",
    "MultiplyValue",
    "RemoveValueRows",
    "normalize"
]



"""
Module to encode the class hierarchy of the global fishing watch
"""
from __future__ import annotations

import re
from typing import Dict, List, Optional, Union


class VesselType:
    """
    The base class for all vessel types defined by the global fishing watch.
    """

    _all_types: Optional[List[VesselType]] = None

    @classmethod
    def typename(cls) -> str:
        """
        Get the representation name.

        :return: The typename in lower case and snake case
        """
        snake_case_name = cls.__name__
        snake_case_name = re.sub('([A-Z]+)', r'_\1', snake_case_name).lower()
        snake_case_name = re.sub('^_', '', snake_case_name)
        return snake_case_name

    @classmethod
    def get_types(cls) -> List[VesselType]:
        """
        Get all available vessel types.

        :return: List of vessel types
        """
        klasses = []
        for subclass in cls.__subclasses__():
            klasses.append(subclass)
            klasses.extend(cls._subclasses(subclass))
        return klasses

    @classmethod
    def get_types_as_str(cls) -> List[VesselType]:
        return [x.typename() for x in cls.get_types()]

    @staticmethod
    def _subclasses(cls) -> List[VesselType]:
        klasses = []
        for subclass in cls.__subclasses__():
            klasses.append(subclass)
            klasses.extend(cls._subclasses(subclass))
        return klasses

    @classmethod
    def get_values(cls) -> List[int]:
        """
        Get the int representations for this class

        :return: List of values
        """
        cls._initialize_types()
        values: List[int] = []

        for klass in cls._all_types:
            values.append(VesselType.to_id(klass=klass))

        return values

    @classmethod
    def _initialize_types(cls):
        if cls._all_types is None:
            cls._all_types = cls.get_types()

    @classmethod
    def by_id(cls, *,
              identifier: int) -> VesselType:
        cls._initialize_types()

        return cls._all_types[identifier]

    @classmethod
    def to_id(cls, *,
              klass: Union[str, VesselType] = None) -> int:
        """
        Get the id for a klass name or class type of VesselType.

        :param klass:
        :return: id for a particular vessel class
        """
        VesselType._initialize_types()
        if klass is None:
            klass = cls

        if type(klass) is str:
            klass = cls.by_name(name=klass)

        if issubclass(klass, VesselType):
            return cls._all_types.index(klass)

        raise ValueError(f"VesselType.by_id: failed to identify '{klass}'")

    @classmethod
    def by_name(cls,
                name: str) -> VesselType:
        """
        Get the VesselType by given name

        :param name: Name (representation) of the type
        :return: VesselType Class Object
        """
        cls._initialize_types()

        for k in cls._all_types:
            if k.typename() == name:
                return k

        raise KeyError(f"VesselType.by_name: failed to identify '{name}'")

    def __class_getitem__(cls, name: str) -> int:
        """
        Allow an Enum-like interface to the class index values

        :param name: Name of the class
        :return: int representation of the class
        """
        klass = cls.by_name(name=name)
        return cls.to_id(klass=klass)

    @classmethod
    def get_mapping(cls) -> Dict[str, int]:
        """
        Compute the mapping from vessel typename to integer

        :return: Dictionary representing the mapping
        """
        mapping = {}
        for t in cls.get_types():
            mapping[t.typename()] = cls.to_id(klass=t)
        return mapping


class Unspecified(VesselType):
    pass


class Cargo(VesselType):
    pass


class Passenger(VesselType):
    pass


class Pleasure(VesselType):
    pass


class Specialcraft(VesselType):
    pass


class Tanker(VesselType):
    pass


class Tug(VesselType):
    pass


class Fishing(VesselType):
    pass

# region Global Fishing Watch Types


class SquidJigger(Fishing):
    pass


class DriftingLonglines(Fishing):
    pass


class PoleAndLine(Fishing):
    pass


class Trollers(Fishing):
    pass


class FixedGear(Fishing):
    pass


class Trawlers(Fishing):
    pass


class DredgeFishing(Fishing):
    pass


class Seiners(Fishing):
    pass


class PurseSeines(Seiners):
    pass


class OtherSeines(Seiners):
    pass


class TunaPurseSeines(PurseSeines):
    pass


class OtherPurseSeines(PurseSeines):
    pass


class PotsAndTraps(FixedGear):
    pass


class SetLonglines(FixedGear):
    pass


class SetGillnets(FixedGear):
    pass

# endregion



"""
Module to define an annotated dataframe, i.e. the combination of data and metadata.
"""
from __future__ import annotations

import copy
import gc
import json
import logging
import os
from logging import INFO, Logger, getLogger
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Callable, List, Optional, Union

import polars
import pyarrow
import pyarrow.parquet as pq
from tqdm import tqdm

from .annotations import Annotation
from .constants import DAMAST_SPEC_SUFFIX, DAMAST_SUPPORTED_FILE_FORMATS
from .data_description import ListOfValues, MinMax
from .metadata import DataSpecification, MetaData, ValidationMode
from .types import DataFrame, XDataFrame

__all__ = ["AnnotatedDataFrame"]


logging.basicConfig()
_log: Logger = getLogger(__name__)
_log.setLevel(INFO)


class AnnotatedDataFrame(XDataFrame):
    """
    A dataframe that is associated with metadata.

    :param dataframe: The polars dataframe holding the data
    :param metadata: The metadata for the dataframe
    :param validation_mode:
        - If :attr:`damast.core.ValidationMode.UPDATE_DATA` replace values outside of
          valid range (specified in :attr:`metadata`) with missing value.
        - If :attr:`damast.core.ValidationMode.UPDATE_METADATA` update the metadata according to the encountered values.
        - Else :attr:`damast.core.ValidationMode.READONLY` will throw when encountering inconsistencies
    """

    #: Metadata associated with the dataframe
    _metadata: MetaData

    #: The actual dataframe
    _dataframe: DataFrame

    def __init__(
        self,
        dataframe: polars.DataFrame | polars.LazyFrame | XDataFrame,
        metadata: MetaData,
        validation_mode: ValidationMode = ValidationMode.READONLY,
    ):
        if isinstance(dataframe, XDataFrame):
            dataframe = dataframe._dataframe

        if isinstance(dataframe, polars.DataFrame):
            dataframe = dataframe.lazy()

        if not isinstance(dataframe, polars.LazyFrame):
            raise ValueError(
                f"{self.__class__.__name__}.__init__: dataframe must be"
                f" of type 'DataFrame', but was '{type(dataframe)}"
            )

        if not isinstance(metadata, MetaData):
            raise ValueError(
                f"{self.__class__.__name__}.__init__: metadata must be"
                f" of type 'MetaData', but was '{type(metadata)}"
            )

        super().__init__(df=dataframe)

        self._metadata = metadata

        # Ensure conformity of the metadata with the dataframe
        self.validate_metadata(validation_mode=validation_mode)

    @property
    def metadata(self) -> MetaData:
        """Get the metadata for this dataframe"""
        return self._metadata

    @classmethod
    def ensure_type(cls, obj: any):
        if not isinstance(obj, cls):
            raise ValueError("Object {obj} is not an AnnotatedDataFrame")

        if not isinstance(obj._dataframe, DataFrame):
            raise ValueError(f"AnnotatedDataFrame._dataframe is not of type {DataFrame}")

    def validate_metadata(
        self, validation_mode: ValidationMode = ValidationMode.READONLY
    ) -> None:
        """
        Validate this annotated dataframe and ensure that data and spec match.

        :param validation_mode: Select the validation mode that should be used
        :raise RuntimeError: Dependending on the validation mode an exception will be raise to ensure the data spec
               conformance
        """
        self._dataframe = self._metadata.apply(df=self._dataframe, validation_mode=validation_mode)

    def is_empty(self) -> bool:
        """
        Check if annotated dataframe has associated data.

        :return: False if there is an internal :code:`_dataframe` set, True otherwise
        """
        return self._dataframe is None

    def get_fulfillment(
        self, expected_specs: List[DataSpecification]
    ) -> MetaData.Fulfillment:
        """
        Get the :class:`MetaData.Fulfillment` with respect to the given expected specification.

        :param expected_specs: The expected specification
        :return: Instance of :class:`MetaData.Fulfillment` to investigate the degree of fulfillment
        """
        return self._metadata.get_fulfillment(expected_specs=expected_specs)

    def update(self, expectations: List[DataSpecification]):
        """
        Update the metadata based on a set of validated expectations.

        :param expectations: List of :class:`DataSpecifications` as expectations that this data meets.
        """
        for expected_data_spec in expectations:
            column_name = expected_data_spec.name
            if column_name not in self.metadata:
                # Column name description is not yet part of the metadata
                # Verify that is it part of the data frame
                if column_name in self.dataframe.column_names:
                    self._metadata.columns.append(
                        DataSpecification.from_dict(data=dict(expected_data_spec))
                    )
                else:
                    raise RuntimeError(
                        f"{self.__class__.__name__}.update:"
                        f" required output '{column_name}' is not"
                        f" present in the result dataframe - available columns are:"
                        f" {','.join(self.dataframe.column_names)}"
                    )


    def save(self, *, filename: Union[str, Path]) -> AnnotatedDataFrame:
        """
        Save this annotated dataframe in a file.

        Filetype can be .parquet (recommended) or *.hdf5.

        For hdf5 the resulting file can be inspected using HDF5 tools and in particular h5dump.
        To get the HDF5 groups and their corresponding metdata (header only e.g. no data printed), use the command:

        .. code-block:: console

            h5dump -H data.hdf5

        See: https://portal.hdfgroup.org/display/HDF5/HDF5+Command-line+Tools

        :param filename: Filename to use for saving

        """
        if self._dataframe is None:
            raise ValueError(f"{self.__class__.__name__}.save: no dataframe to save")

        metadata_filename = Path(filename).with_suffix(DAMAST_SPEC_SUFFIX)
        self._metadata.save_yaml(filename=metadata_filename)

        if Path(filename).suffix not in [".hdf5", ".h5"]:
            self.export(filename)
            return self

        # First save the hdf5 file in order to add then the metadata to it
        XDataFrame.export_hdf5(self._dataframe, filename)
        self._metadata.append_to_hdf(filename)

        return self

    def export(self, filename: str | Path):
        """
        Export the annotated dataframe to a file. By default the format is parquet.
        """
        arrow_table = self._dataframe.compat.collected().to_arrow()
        new_schema = arrow_table.schema.with_metadata({b'annotated_dataframe': json.dumps(dict(self._metadata), default=str).encode('UTF-8')})
        arrow_table = pyarrow.Table.from_arrays(arrow_table.columns, schema=new_schema)
        pq.write_table(arrow_table, filename)

    @classmethod
    def get_supported_format(cls, suffix: str) -> str | None:
        """
        Get the name of the supported format from the suffix
        """
        for format_name, suffixes in DAMAST_SUPPORTED_FILE_FORMATS.items():
            if suffix in suffixes:
                return format_name

        return None

    @classmethod
    def from_files(cls,
            files: list[str|Path],
            metadata_required: bool = True,
            validation_mode: ValidationMode = ValidationMode.READONLY
        ) -> AnnotatedDataFrame:
        """
        Create an annotated dataframe by loading given files

        :param files: Files to use for importing and creating the annotated dataframe

        """
        metadata = None
        if not files or type(files) != list:
            raise ValueError(f"{cls.__name__}.from_files: file list required, but were {files=}")

        suffixes = set()
        for f in files:
            path = Path(f)
            if not path.exists():
                raise FileNotFoundError(f"{cls.__name__}.from_files: could not find {f}")

            suffixes.add(path.suffix)

        if len(suffixes) != 1:
            raise RuntimeError(f"{cls.__name__}.from_files: one file type expected, but multiple suffixes found: {suffixes=}")

        suffix = list(suffixes)[0]
        load_fn = None
        # find load function by suffix
        for filetype, suffixes in DAMAST_SUPPORTED_FILE_FORMATS.items():
            if suffix in suffixes:
                if hasattr(cls, f"load_{filetype}"):
                    load_fn = getattr(cls, f"load_{filetype}")

        if load_fn:
            df, metadata = load_fn(files)
        else:
            raise RuntimeError(
                f"Could not load {files} - supported are currently {DAMAST_SUPPORTED_FILE_FORMATS}"
                " where .h5, .hdf5 generated by vaex or pandas"
            )

        if not metadata:
            _log.info("No metadata provided or found in files - searching now for an existing spec file")
            metadata, _ , metadata_file_candidates = MetaData.search(files)

            if metadata is None:
                _log.info("Found no candidate for a spec file")
                if metadata_required:
                    _log.info("Metadata is required but not available")
                    head = df.head(10).collect()
                    raise RuntimeError(
                        f"{cls.__name__}.from_files:"
                        f" metadata is missing for {files=}"
                        f" and needs to be added'\n"
                        f"{head} - {[str(x) for x in metadata_file_candidates]}"
                    )
                else:
                    _log.info("Metadata is not available and not required, so inferring annotation")
                    metadata = cls.infer_annotation(df)
                    _log.info("Metadata inferring completed")
                    return cls(dataframe=df, metadata=metadata, validation_mode=ValidationMode.IGNORE)
            else:
                _log.info(f"Found metadata: {[str(x) for x in metadata_file_candidates]}")
        elif len(metadata) != len(files):
            _log.info("Metadata is not available for all files, so inferring annotation")
            metadata = cls.infer_annotation(df)
            _log.info("Metadata inferring completed")
            return cls(dataframe=df, metadata=metadata, validation_mode=ValidationMode.IGNORE)
        elif len(metadata) == len(files):
            metadata_list = list(metadata.values())
            metadata = MetaData(
                        columns=metadata_list[0].columns,
                        annotations=list(metadata_list[0].annotations.values())
                    )
            for i in range(0, len(metadata_list)-1):
                metadata = metadata.merge(metadata_list[i+1])
        else:
            for _, m in metadata.items():
                metadata = m
                break

        return cls(dataframe=df, metadata=metadata, validation_mode=validation_mode)

    @classmethod
    def load_parquet(cls, files) -> tuple[AnnotatedDataFrame, dict[str, MetaData]]:
            _log.info(f"Loading parquet: {files=}")
            metadata = None
            df = polars.scan_parquet(files)

            metadata_per_file = {}
            for file in files:
                schema = pq.read_schema(file)
                if schema and hasattr(schema, "metadata"):
                    data = schema.metadata[b"annotated_dataframe"]
                    m = MetaData.from_dict(json.loads(data.decode('UTF-8')))
                    m.set_annotation(Annotation(name=Annotation.Key.Source, value=Path(file).name))
                    metadata_per_file[file] = m
            return df, metadata_per_file

    @classmethod
    def load_netcdf(cls, files) -> tuple[AnnotatedDataFrame, dict[str, MetaData]]:
        _log.info(f"Loading netcdf: {files=}")
        return XDataFrame.import_netcdf(files)

    @classmethod
    def load_hdf(cls, files) -> tuple[AnnotatedDataFrame, dict[str, MetaData]]:
        _log.info(f"Loading hdf: {files=}")
        return XDataFrame.import_hdf5(files)

    @classmethod
    def load_csv(cls, files) -> tuple[AnnotatedDataFrame, dict[str, MetaData]]:
        _log.info(f"Loading csv: {files=}")
        df = polars.scan_csv(files, separator=";", quote_char=None, infer_schema_length=None)
        if len(df.compat.column_names) <= 1:
            # unlikely that this frame has only one column, so trying with comma
            df = polars.scan_csv(files, separator=",", quote_char=None, infer_schema_length=None)
        return df, {}

    @classmethod
    def from_file(cls,
            filename: str | Path,
            metadata_required: bool = True,
        ) -> AnnotatedDataFrame:
        """
        Create an annotated dataframe from an hdf5 file.

        :param filename: Filename to use for importing and creating the annotated dataframe

        """
        return cls.from_files([filename], metadata_required=metadata_required)

    @classmethod
    def infer_annotation(cls, df: DataFrame) -> MetaData:
        column_specs: list[DataSpecification] = []

        numeric_columns: list[str] = []
        for column in tqdm(df.compat.column_names, desc=f"Extract str and categorical column metadata"):
            data = {'name': column,
                    'is_optional': False,
                    'representation_type': df.compat.dtype(column)
            }

            if df.compat.is_string(column):
                categories = df.compat.categories(column)
                if categories:
                    data['value_range'] = ListOfValues(categories)
            elif df.compat.is_numeric(column):
                numeric_columns.append(column)
                continue

            if 'value_range' not in data:
                min_value, max_value = df.compat.minmax(column)
                if min_value is not None and max_value is not None:
                    data['value_range'] = MinMax(min_value, max_value)

            ds = DataSpecification(**data)
            column_specs.append(ds)

        if numeric_columns:
            # To allow polars to optimize the query, process all numeric columns at once
            results = df.compat.minmax_stats(numeric_columns)
            for column in tqdm(numeric_columns, desc="Extract numeric column metadata"):
                data = {'name': column,
                        'is_optional': False,
                        'representation_type': df.compat.dtype(column)
                }

                min_value = results[column]["min_value"]
                max_value = results[column]["max_value"]
                if min_value is not None and max_value is not None:
                    data['value_range'] = MinMax(results[column]["min_value"], results[column]["max_value"])
                data['value_stats'] = results[column]["stats"]

                ds = DataSpecification(**data)
                column_specs.append(ds)

        return MetaData(columns=column_specs, annotations=[])

    @classmethod
    def convert_csv_to_adf(
        cls,
        csv_filenames: List[Union[Path, str]],
        metadata_filename: Union[Path, str],
        output_filename: Union[Path, str],
        validation_mode: ValidationMode = ValidationMode.READONLY,
        progress: Optional[Callable[[float], None]] = None,
        csv_sep: str = ";",
    ):
        """
        Convert a csv file to an annotated dataframe

        :param csv_filenames: The input csv file that shall be converted
        :param metadata_filename: The metadata specification
        :param output_filename: The output file that will be generated
        :param validation_mode: how to validate / enforce the conformity of the metadata and the data
        :param progress: Callable to set for the dataframe conversion
        :param csv_sep: Separator to use when loading csv files
        """
        metadata = MetaData.load_yaml(filename=metadata_filename)
        df = polars.scan_csv(sorted(csv_filenames),
                            separator=csv_sep,
                            quote_char=None,
                            infer_schema_length=None
            )
        adf = cls(dataframe=df, metadata=metadata)

        _log.info(f"Metadata: {dict(metadata)}")
        _log.info(f"Saving dataframe into {output_filename}")

        adf.save(filename=output_filename)

    def drop(self, columns, strict: bool = True) -> AnnotatedDataFrame:
        self._dataframe = self._dataframe.drop(columns, strict=strict)
        self._metadata.drop(columns)
        return self

    def copy(self):
        return copy.deepcopy(self)

    @property
    def shape(self):
        return self._dataframe.compat.collected().shape

    def __deepcopy__(self, memo=None):
        return AnnotatedDataFrame(self._dataframe.clone(), copy.deepcopy(self._metadata))



from argparse import ArgumentParser
from pathlib import Path

from damast.cli.base import BaseParser


class ExperimentParser(BaseParser):
    """
    Argparser for handling experiments

    :param parser: The base parser
    """

    def __init__(self, parser: ArgumentParser):
        super().__init__(parser=parser)

        parser.description = "damast experiment - allow to execute experiments"
        parser.add_argument("-f", "--filename",
                            help="Filename of the experiment that should be executed",
                            required=True
                            )
        parser.add_argument("-o", "--output-dir", dest="output_dir",
                            default=str((Path() / "output").resolve()),
                            help="Absolute path to output folder",
                            required=True)

    def execute(self, args):
        super().execute(args)

        from damast.ml.experiments import Experiment

        experiment = Experiment.from_file(args.filename)
        experiment.output_directory = args.output_dir

        experiment.run(logging_level=args.loglevel)



"""
Module to define Annotation object for data specification
"""
# Copyright (C) 2023-2025 Simula Research Laboratory
#
# This file is part of damast
#
# SPDX-License-Identifier:    BSD-3-Clause
from __future__ import annotations

from datetime import datetime
from enum import Enum
from typing import Any, ClassVar, Dict, List, Optional, Union

__all__ = ["Annotation", "Change", "History"]


class Annotation:
    """
    Annotation class

    :param name: Name / Type of the annotation (see also :class:`Key`)
    :param value: Value of the annotation
    """

    class Key(str, Enum):
        """
        Enumeration of predefined keys which can be used for the annotation
        """

        Institution = "institution"
        License = "license"
        References = "references"
        Source = "source"
        History = "history"
        Comment = "comment"

    #: Name of the annotation (see also Key)
    name: str

    #: Value of the annotation
    value: Optional[Any] = None

    def __init__(self, name: Union[str, Key], value: Optional[Any] = None):
        self.name = name
        self.value = value

        # In case we want to enforce some key-related validation, then
        # add a 'validate_<key-name>' function
        if isinstance(self.name, Annotation.Key):
            self.name = self.name.value

        validation_func_name = f"validate_{self.name}"
        if hasattr(self, validation_func_name):
            getattr(self, validation_func_name)()

    def __eq__(self, other) -> bool:
        """
        Check if two annotations are the same.

        The two objects are considered equal if the other object is a :class:`Annotation`
        is the same for both objects, and the  :py:attr:`name` and :py:attr:`value` of
        the annotation is the same.

        :param other: Other object
        :return: ``True`` if objects are consider same, otherwise ``False``
        """
        if self.__class__ != other.__class__:
            return False

        if self.value != other.value:
            return False

        if self.name != other.name:
            return False

        return True

    def __iter__(self):
        yield self.name, self.value

    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> Annotation:
        """
        Create an instance from a given dictionary.

        .. note::
            The dictionary is assumed to only have one ``(key, value)`` pair,
            where the ``key`` is interpreted as the :attr:`Annotation.name`, the
            ``value`` as the :attr:`Annotation.value`

        :param data: The input dictionary
        :return: Annotation
        :raise RuntimeError: Raises if annotation cannot be created from dictionary
        """
        assert len(data.keys()) == 1, "Input dictionary has more than one object"
        for k, value in data.items():
            return Annotation(name=k, value=value)

    def validate_license(self) -> None:
        """
        Validation function for the license

        :raise ValueError: Raises if a license is given, but it cannot be validated
        """
        if self.value is None or self.value == "":
            raise ValueError("License cannot be empty")

    def validate_comment(self):
        """
        Validation function for the comment

        :raise ValueError: Raises if a comment remains empty.
        """
        if self.value is None or self.value == "":
            raise ValueError("Comment cannot be empty")


class Change:
    """
    Representation of a change of data and metadata by transformation pipelines

    :param title: A concise description of the change
    :param description: A detailed description of the change
    :param timestamp: A timestamp for when the change was mede
    """

    TIMESTAMP_FORMAT: ClassVar[str] = "%Y-%m-%d %H:%M:%S"

    title: str
    timestamp: datetime
    description: str

    def __init__(
        self, title: str, description: str, timestamp: datetime = datetime.utcnow()
    ):
        # Ensure that we only deal with the required precision
        timestamp_txt = timestamp.strftime(self.TIMESTAMP_FORMAT)
        self.timestamp = datetime.strptime(timestamp_txt, self.TIMESTAMP_FORMAT)

        self.title = title
        self.description = description

    def __repr__(self):
        return f"{self.timestamp.strftime(self.TIMESTAMP_FORMAT)}: {self.title}"

    def __eq__(self, other):
        if self.__class__ != other.__class__:
            return False

        if self.title != other.title:
            return False
        if self.timestamp != other.timestamp:
            return False
        if self.description != other.description:
            return False

        return True

    def __iter__(self):
        yield "title", self.title
        yield "timestamp", self.timestamp.strftime(self.TIMESTAMP_FORMAT)
        yield "description", self.description

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """Initialize a change from a dictionary

        .. note::
            The dictionary is required to have the keys ``"title"``,
            ``"timestamp"`` and ``"description"``.
            All other keys in the input dictionary is ignored

        :param data: Input dictionary
        :return: A change object
        """
        return cls(
            title=data["title"],
            timestamp=datetime.strptime(
                data["timestamp"], Change.TIMESTAMP_FORMAT
            ),
            description=data["description"],
        )


class History(Annotation):
    """
    Representation of the (linear) history of changes

    :param changes: A list of changes (ordered chronologically)
    """

    changes: List[Change]

    def __init__(self, changes: Optional[List[Change]] = None):
        super().__init__(name=Annotation.Key.History)
        if changes is None:
            self.changes = []
        else:
            self.changes = sorted(changes, key=str)

    def __eq__(self, other) -> bool:
        """
        A history is equal to another history of it has the same sequence of changes

        :param other: The other object
        """
        if self.__class__ != other.__class__:
            return False

        return self.changes == other.changes

    def add_change(self, change: Change):
        """
        Extend history with adding a new change

        :param change: The change
        """
        changes = self.changes + [change]
        self.changes = sorted(changes, key=str)

    def __iter__(self):
        yield Annotation.Key.History.value, [dict(c) for c in self.changes]

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        """
        Initialize a history from a dictionary

        .. note::
            It is assumed that the history is located under
            the key (string representation) of :attr:`Annotation.Key.History`.
            The item under this key should be a list changes (represented as dictionaries).

        .. note::
            If the key  :attr:`Annotation.Key.History` is not present in the dictionary, assume no
            changes present and initialize a clean history

        :param data: The input dictionary
        :return: A history object with the corresponding changes
        """
        changes = []
        if Annotation.Key.History.value in data:
            for c in data[Annotation.Key.History.value]:
                changes.append(Change.from_dict(data=c))

        return cls(changes=changes)



"""
Module containing functionality to setup up an Machine-Learning Experiment
"""
from __future__ import annotations

import datetime
import importlib
import os
import random
import tempfile
from logging import INFO, Logger, basicConfig, getLogger
from pathlib import Path
from typing import Any, Dict, List, NamedTuple, Optional, Sequence, Union

import keras
import numpy as np
import polars as pl
import yaml

from damast.core.dataframe import AnnotatedDataFrame
from damast.core.dataprocessing import DataProcessingPipeline
from damast.core.types import DataFrame
from damast.data_handling.accessors import GroupSequenceAccessor
from damast.ml.models.base import BaseModel, ModelInstanceDescription

basicConfig()
_log: Logger = getLogger(__name__)

__all__ = [
    "Experiment",
    "ForecastTask",
    "LearningTask",
    "ModelInstanceDescription",
    "TrainingParameters"
]


class TrainingParameters(NamedTuple):
    epochs: int = 10
    steps_per_epoch: int = 10
    validation_steps: int = 12
    learning_rate: float = 0.1
    loss_function: str = "mse"


class LearningTask:
    """
    Description of a trainable Learning Task.

    This class can associate a processing pipeline with a set of model to train and evaluation.

    :param label: Label of this learning task (e.g. to uniquely mark the combination of task and learning parameters)
           This label will be used to suffix the model-specific subfolder
    :param pipeline: The processing / feature extraction pipeline
    :param features: The features that this machine-learning model requires as input
    :param targets: The output features aka targets that this machine-learning model provides
    :param models: The actual model instances, i.e. model + parameters, that this learning
           task shall comprise
    """
    label: str
    pipeline: DataProcessingPipeline
    features: List[str]
    targets: List[str]

    models: List[ModelInstanceDescription]
    training_parameters: Union[Dict[str, Any], TrainingParameters] = TrainingParameters(),

    def __init__(self, *,
                 label: str,
                 pipeline: Union[Dict[str, Any], DataProcessingPipeline],
                 features: List[str],
                 models: List[Union[Dict[str, Any], ModelInstanceDescription]],
                 targets: Optional[List[str]] = None,
                 training_parameters: Union[Dict[str, Any], TrainingParameters] = TrainingParameters(),
                 ):

        self.label = label

        if isinstance(pipeline, DataProcessingPipeline):
            self.pipeline = pipeline
        elif isinstance(pipeline, dict):
            self.pipeline = DataProcessingPipeline(**pipeline)
        else:
            raise ValueError(f"{self.__class__.__name__}.__init__: could not instantiate DataProcessingPipeline"
                             f" from {type(pipeline)}")

        self.features = features
        if targets is None:
            self.targets = self.features
        else:
            self.targets = targets

        self.models = []
        for m in models:
            if isinstance(m, ModelInstanceDescription):
                self.models.append(m)
            elif isinstance(m, dict):
                self.models.append(ModelInstanceDescription.from_dict(data=m))
            else:
                raise ValueError(f"{self.__class__.__name__}.__init__: could not instantiate ModelInstanceDescription"
                                 f" from {type(m)}")

        if isinstance(training_parameters, TrainingParameters):
            self.training_parameters = training_parameters
        elif isinstance(training_parameters, dict):
            self.training_parameters = TrainingParameters(**training_parameters)
        else:
            raise ValueError(f"{self.__class__.__name__}.__init__: training_parameters must be either "
                             f"dict or TrainingParameters object")

    @classmethod
    def from_dict(cls, data: Dict[str, Any]):
        if "module_name" not in data:
            raise KeyError(f"{cls.__name__}.create: missing 'module_name'")

        if "class_name" not in data:
            raise KeyError(f"{cls.__name__}.create: missing 'class_name'")

        module_name = data["module_name"]
        class_name = data["class_name"]

        m_module = importlib.import_module(module_name)
        if hasattr(m_module, class_name):
            del data["module_name"]
            del data["class_name"]

            klass = getattr(m_module, class_name)
            return klass(**data)

        raise ValueError(f"{cls.__name__}.from_dict: could not find '{class_name}' in '{module_name}'")

    def __iter__(self):
        yield "module_name", self.__class__.__module__
        yield "class_name", self.__class__.__qualname__

        yield "label", self.label
        yield "pipeline", dict(self.pipeline)
        yield "features", self.features
        yield "targets", self.targets
        yield "models", [dict(x) for x in self.models]

        yield "training_parameters", self.training_parameters._asdict() # type: ignore

    def __eq__(self, other) -> bool:
        if type(self) != type(other):
            return False

        for attr in ["pipeline", "features", "targets", "models"]:
            if getattr(self, attr) != getattr(other, attr):
                return False
        return True


class ForecastTask(LearningTask):
    sequence_length: int
    forecast_length: int
    group_column: str

    def __init__(self, *,
                 label: str,
                 pipeline: Union[Dict[str, Any], DataProcessingPipeline],
                 group_column: str,
                 features: List[str],
                 sequence_length: int,
                 forecast_length: int,
                 models: List[Union[Dict[str, Any], ModelInstanceDescription]],
                 targets: Optional[List[str]] = None,
                 training_parameters: Union[Dict[str, Any], TrainingParameters] = TrainingParameters(),
                 ):
        super().__init__(label=label,
                         pipeline=pipeline,
                         features=features,
                         models=models,
                         targets=targets,
                         training_parameters=training_parameters)

        self.group_column = group_column
        self.sequence_length = sequence_length
        self.forecast_length = forecast_length

    def __eq__(self, other):
        if type(self) != type(other):
            return False

        for attr in ["pipeline", "features", "targets", "models", "group_column", "sequence_length",
                     "forecast_length", "training_parameters"]:
            if getattr(self, attr) != getattr(other, attr):
                return False

        return True

    def __iter__(self):
        for x in super().__iter__():
            yield x

        yield "group_column", self.group_column
        yield "sequence_length", self.sequence_length
        yield "forecast_length", self.forecast_length


class Experiment:
    MARKER_FILE: str = ".damast_experiment"
    TIMESTAMP_FORMAT: str = "%Y%m%d-%H%M%S"

    learning_task: LearningTask

    input_data: Path
    output_directory: Path
    label: str

    _batch_size: int
    _split_data_ratios: List[float]

    _timestamp: datetime.datetime
    _evaluation_steps: int
    _evaluation_report: Dict[str, Dict[str, Any]]

    _trained_models: List[BaseModel]

    def __init__(self,
                 learning_task: Union[Dict[str, Any], LearningTask],
                 input_data: Union[str, Path],
                 output_directory: Union[str, Path] = tempfile.gettempdir(),
                 batch_size: int = 2,
                 evaluation_steps=1,
                 split_data_ratios: List[float] = [1.6, 0.2, 0.2],
                 label: str = "damast-ml-experiment",
                 timestamp: Union[str, datetime.datetime] = datetime.datetime.utcnow(),
                 evaluation={}
                 ):
        """
        The ratios of how to split (test, training, validation) data from the input dataset

        :param learning_task:  A description of the machine-learning model setup
        :param input_data:
        :param evaluation_steps:
        :param split_data_ratios:
        :param label:
        :param timestamp:
        :param evaluation:
        :param batch_size: The batch-size
        :param output_directory: Path to direct output to
        """
        if isinstance(learning_task, LearningTask):
            self.learning_task = learning_task
        elif isinstance(learning_task, dict):
            self.learning_task = LearningTask.from_dict(data=learning_task)
        else:
            raise ValueError(f"{self.__class__.__name__}.__init__: learning_task must be either"
                             f"dict or LearningTask object")

        if type(input_data) == list:
            self.input_data = input_data
        else:
            self.input_data = [ Path(input_data) ]

        self.output_directory = Path(output_directory)
        self.label = label

        self._batch_size = batch_size
        self._evaluation_steps = evaluation_steps
        self._split_data_ratios = split_data_ratios
        if isinstance(timestamp, str):
            timestamp = datetime.datetime.strptime(timestamp, Experiment.TIMESTAMP_FORMAT)

        self._timestamp = timestamp
        self._evaluation_report = evaluation
        self._trained_models = []

    @classmethod
    def from_file(cls,
                  filename: Union[str, Path]) -> Experiment:
        if not Path(filename).exists():
            raise FileNotFoundError(f"{cls.__name__}.from_file: the given experiment description file"
                                    f" does not exist: '{filename}'")

        with open(filename, "r") as f:
            data = yaml.load(f, Loader=yaml.SafeLoader)
            return cls(**data)

    def save(self,
             filename: Union[str, Path]):
        """
        Save the experiment as yaml.

        :param filename: Name of the file
        """
        with open(filename, "w") as f:
            yaml.dump(dict(self), f)

    @classmethod
    def create_experiment_directory(cls,
                                    base_dir: Union[str, Path],
                                    label: str) -> Path:
        """Create an experiment directory inside the output directory.

        The created directory is prefixed with a time-stamp indicating when it was created.

        """
        prefix = datetime.datetime.utcnow().strftime(cls.TIMESTAMP_FORMAT)
        experiment_dir = base_dir / f"{prefix}-{label}"
        experiment_dir.mkdir(exist_ok=False, parents=True)

        Experiment.touch_marker(experiment_dir)

        return experiment_dir

    @classmethod
    def validate_experiment_dir(cls, dir: Union[str, Path]) -> Path:
        """
        Validate that a particular directory is an "experiment" directory.

        Experiment directories will have a hidden marker file

        :param dir: Path or name of the directorry
        :return: Directory Path
        """
        experiment_dir = Path(dir)
        if not experiment_dir.exists():
            raise FileNotFoundError(f"{cls.__name__}.from_directory: {dir} does not exist")

        if not experiment_dir.is_dir():
            raise NotADirectoryError(f"{cls.__name__}.from_directory: {dir}")

        if not (experiment_dir / cls.MARKER_FILE).exists():
            raise NotADirectoryError(f"{cls.__name__}.from_directory: {dir} is not an experiment result directory")

        return experiment_dir

    @classmethod
    def from_directory(cls, dir: Union[str, Path]) -> Dict[str, 'keras.Model']:
        """
        Create an experiment object by loading a directory with experiment artifacts.

        Note that currently keras models will be loaded, when available.

        :param dir: Directory with experiment artifacts
        :return: Dictionary of loaed
        """
        experiment_dir = cls.validate_experiment_dir(dir=dir)

        models = {}

        import keras.models

        from damast.ml.models.base import MODEL_TF_HDF5

        # Load available models
        for dirname in os.listdir(str(experiment_dir)):
            model_hdf5 = experiment_dir / dirname / MODEL_TF_HDF5
            if model_hdf5.exists():
                models[Path(dirname).stem] = keras.models.load_model(model_hdf5)

        return models

    @classmethod
    def touch_marker(cls, dir: Union[str, Path]):
        """
        Create a marker file in a given directory.

        :param dir: Directory where to create the marker file
        """
        path = Path(dir)
        if not path.exists():
            raise FileNotFoundError(f"{cls.__name__}.touch_marker: {path} does not exist")
        elif not path.is_dir():
            raise NotADirectoryError(f"{cls.__name__}.touch_marker: {path} is not a directory")

        with open(path / cls.MARKER_FILE, "a"):
            pass

    @classmethod
    def compute_train_test_validate_groups(cls,
                                           adf: AnnotatedDataFrame,
                                           group: str,
                                           ratios: List[float]) -> List[List[Any]]:
        """
        Given a :class:`damast.AnnotatedDataFrame` and a column name, split the dataframe into groups
        based on the column name. The size of each group is determined by the ratio.

        ..note::
            This function does not modify anything in the incoming dataframe.

        :param adf: The available data in the input data frame
        :param group: In case this deals with timeseries / sequence data, the group will relate to multiple rows
        :param ratios: Ratio between the train, test and validate dataset
        """

        normalized_rates = np.asarray([rate / sum(ratios) for rate in ratios])
        groups = adf[group].unique().select(pl.col(group).shuffle().alias(group)).collect()

        partition_sizes = np.asarray(
            np.round(len(groups) * normalized_rates), dtype=int)
        delta = len(groups) - sum(partition_sizes)
        assert delta <= 1, f"|# of groups {len(groups)} - # of partitions {sum(partition_sizes)}| <= 1"
        if delta == 1:
            rand_idx = random.randint(0, len(partition_sizes) - 1)
            partition_sizes[rand_idx] += 1

        if len(groups) != sum(partition_sizes):
            raise RuntimeError(f"Expected group_len == sum partition_sizes, but was {len(groups)} == {sum(partition_sizes)}")

        from_idx = 0
        partitions = []
        for ps in partition_sizes:
            to_idx = min(from_idx + ps, len(groups))
            partitions.append(groups[from_idx:to_idx])
            from_idx = to_idx
        return partitions

    def compute_features(self,
                         adf: AnnotatedDataFrame) -> AnnotatedDataFrame:
        """
        Compute features for an experiment by running the designated pipeline on input dataframe.

        .. note::
            Stores the transformed data in the experiment directory.

        :param adf: The input dataframe
        """
        adf_with_features = self.learning_task.pipeline.transform(adf)

        self.learning_task.pipeline.save(dir=self.output_directory)
        self.learning_task.pipeline.save_state(
            df=adf_with_features,
            dir=self.output_directory)
        return adf_with_features

    def create_generator(self,
                         adf: AnnotatedDataFrame,
                         group: str,
                         group_ids: List[Any]) -> Sequence:
        """
        Create a generator applicable for :py:mod:`keras` from the input data.

        ..note::
            This function is called whenever you want to train your models, i.e
            :func:`Experiment.train`

        :param adf: The input dataframe
        :param group: Group to extract sequences from
        :param group_ids: Groups used in the generator
        :returns: A triplet of (train, test, validation) generators of the input features
        """

        sta = GroupSequenceAccessor(df=adf,
                                    group_column=group)
        if isinstance(self.learning_task, ForecastTask):
            return sta.to_keras_generator(features=self.learning_task.features,
                                          target=self.learning_task.targets,
                                          sequence_length=self.learning_task.sequence_length,
                                          sequence_forecast=self.learning_task.forecast_length,
                                          groups=group_ids,
                                          batch_size=self._batch_size,
                                          infinite=True)
        else:
            raise NotImplementedError(
                f"{self.__class__.__name__}.create_generator: Sorry, but there is currently no support "
                f"implemented for {self.learning_task.__class__.__name__}")

    def train(self,
              model_instance_description: ModelInstanceDescription,
              train_generator: Sequence,
              validate_generator: Sequence,
              output_dir: Path,
              suffix: str = "",
              epochs: int = 2,
              steps_per_epoch: int = 10,
              validation_steps: int = 12,
              learning_rate: float = 0.1,
              loss_function: str = "mse") -> BaseModel:
        """
        Train a set of machine-learning models on given input data.
        :param suffix: allow to add a training specific suffix to the output folder (otherwise) only the model class name will be used, so only the last training would be kept
        :param model_instance_description: The model instance description for the model that should be trained
        :param train_generator: Generator providing data from training set
        :param validate_generator: Generator providing data from validation set
        :param output_dir: Directory where the model training log, etc. should go
        :param epochs: Number of epochs
        :param steps_per_epoch: Steps per epoch
        :param validation_steps: Number of validation steps
        :param learning_rate: The learning rate for the Stochastic Gradient descent algorithm
        :param loss_function: The loss function

        :return: model instance
        """
        # NOTE: This should probably be an individual step, as we could re-use an existing model to continue training
        if isinstance(self.learning_task, ForecastTask):
            # TODO: the model needs to be applicable to a forecast task - so we might
            model: BaseModel = model_instance_description.model(
                name=f"{model_instance_description.model.__name__}{suffix}",
                output_dir=output_dir,
                features=self.learning_task.features,
                targets=self.learning_task.targets,
                timeline_length=self.learning_task.sequence_length,
                **model_instance_description.parameters
            )
        else:
            raise NotImplementedError(
                f"{self.__class__.__name__}.create_generator: Sorry, but there is currently no support "
                f"implemented for {self.learning_task.__class__.__name__}")

        model.build(loss_function=loss_function,
                    optimizer=keras.optimizers.SGD(learning_rate=learning_rate))

        # Train model
        model.train(training_data=train_generator,
                    validation_data=validate_generator,
                    epochs=epochs,
                    steps_per_epoch=steps_per_epoch,
                    validation_steps=validation_steps)
        model.save()
        return model

    def evaluate(self,
                 test_generator: Sequence,
                 steps: int = 1) -> Dict[str, Dict[str, Any]]:

        return Experiment.evaluate_models(models=self._trained_models,
                                          test_generator=test_generator,
                                          steps=steps)

    @classmethod
    def evaluate_models(cls,
                        models: List[BaseModel],
                        test_generator: Sequence,
                        steps: int = 1) -> Dict[str, Dict[str, Any]]:
        """
        Evaluate a list of trained models.

        :param models: List of trained models
        :param test_generator:  The generator providing access to the test data
        :param steps: number of steps that should be used to test the models

        :return: a dictionary mapping the model name to the results, e.g.
                { "Baseline": { "mse": 0.1, "loss": 0.2 } }
        """
        assert models is not None and len(models) > 0

        evaluations_results = {}
        for model in models:
            evaluation_result = model.evaluate(label=model.name,
                                               evaluation_data=test_generator,
                                               steps=steps)
            evaluations_results[model.name] = evaluation_result
        return evaluations_results

    def run(self,
            logging_level: int = INFO,
            report_filename: Optional[Union[str, Path]] = None) -> DataFrame:
        """
        Run the experiment and return evaluation data.

        The data will also be written into a report file into the current experiment folder unless
        specified otherwise.

        :param logging_level: The logging level that should be applied when running this experiment
        :param report_filename: The report will be written into the experiment folder as experiment dictionary including
            and an additional "evaluation" key

        :return: the evaluation data
        """
        _log.setLevel(logging_level)

        # Load data and validate it by loading it into the annotated dataframe
        # We remove values that are outside the Min/Max value defined in the specification
        adf = AnnotatedDataFrame.from_files(self.input_data)

        if not isinstance(self.learning_task, ForecastTask):
            raise NotImplementedError(
                f"{self.__class__.__name__}.run: Sorry, but there is currently only support "
                f"implemented for ForecastTask")

        group_column = self.learning_task.group_column
        if group_column not in adf.column_names:
            raise RuntimeError(f"{self.__class__.__name__}.run: no column '{group_column}' in the given dataset "
                               f"'{self.input_data}' -- found only '{','.join(adf.column_names)}'")

        # If a learning task is used, that requires a sequence length, then filter the data, so that
        # only relevant data (with a minimum length of the given sequence length) will be accounted
        # for this learning task
        if hasattr(self.learning_task, "sequence_length"):
            groups_with_sequence_length = adf.dataframe.group_by(group_column).agg(sequence_length=pl.len())

            filtered_groups = groups_with_sequence_length.filter(pl.col("sequence_length") > self.learning_task.sequence_length)
            permitted_values = filtered_groups.select(group_column).unique().collect()[:,0]
            adf._dataframe = adf.filter(pl.col(group_column).is_in(permitted_values.implode()))

        features = self.compute_features(adf)
        train_group, test_group, validate_group = \
            self.compute_train_test_validate_groups(features,
                                                    group=group_column,
                                                    ratios=self._split_data_ratios)

        experiment_dir = self.create_experiment_directory(base_dir=self.output_directory,
                                                          label=self.label)

        self._trained_models = []
        for model_instance_description in self.learning_task.models:
            train_data_gen = self.create_generator(features,
                                                   group=group_column,
                                                   group_ids=train_group)
            validate = self.create_generator(features,
                                             group=group_column,
                                             group_ids=validate_group)

            model = self.train(
                suffix=f"-{self.learning_task.label}",
                model_instance_description=model_instance_description,
                train_generator=train_data_gen,
                validate_generator=validate,
                output_dir=experiment_dir,
                **self.learning_task.training_parameters._asdict()  # noqa
            )
            self._trained_models.append(model)

        # TODO: This should return a 'REPORT / PROTOCOL' of the whole experiment
        test_data_gen = self.create_generator(features,
                                              group=group_column,
                                              group_ids=test_group)

        self._evaluation_report = self.evaluate(test_generator=test_data_gen,
                                                steps=self._evaluation_steps)

        self._timestamp = datetime.datetime.utcnow()

        if report_filename is None:
            filename = experiment_dir / "experiment-report.yaml"
        else:
            filename = report_filename

        self.save(filename=filename)
        return filename

    def __eq__(self, other):
        return dict(self) == dict(other)

    def __iter__(self):
        yield "label", self.label
        yield "input_data", [str(x) for x in self.input_data]
        yield "output_directory", str(self.output_directory)
        yield "learning_task", dict(self.learning_task)
        yield "batch_size", self._batch_size
        yield "evaluation_steps", self._evaluation_steps
        yield "split_data_ratios", self._split_data_ratios
        yield "timestamp", self._timestamp.strftime(self.TIMESTAMP_FORMAT)
        yield "evaluation", self._evaluation_report



# -----------------------------------------------------------
# This file contains multiple mathematical function to normalise data, compute distance
#
# (C) 2020 Pierre Bernabé, Oslo, Norway
# email pierbernabe@simula.no
# -----------------------------------------------------------

import numpy as np
import numpy.typing as npt
from numba import njit

__all__ = ["angle_sat_c",
           "great_circle_distance",
           "bearing",
           "reverse_bearing",
           "decdeg2dms",
           "dms2decdeg",
           "distance_sat_vessel",
           "chord_distance"]

EARTH_RADIUS: int = 6371  # km


@njit
def bearing(lat_1: npt.NDArray[np.float64],
            lon_1: npt.NDArray[np.float64],
            lat_2: npt.NDArray[np.float64],
            lon_2: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """
    Computes the bearing/heading between two points.

    The points are defined as
    :math:`(\\phi_1, \\lambda_1)`, :math:`(\\phi_2, \\lambda_2)`
    where :math:`\\phi_i` is the longitude, :math:`\\lambda_i` is the latitude of the points.

    Bearing is defined as the angle between the north-south line of the earth and the line connecting the target
    and the reference point.
    Heading is the angle you are currently navigating in.
    See: https://www.igismap.com/formula-to-find-bearing-or-heading-angle-between-two-points-latitude-longitude/

    :param lat_1: A sequence of latitudes (in degrees), :math:`\\phi_1`
    :param lon_1: A sequence of longitudes (in degrees), :math:`\\lambda_1`
    :param lat_2: A sequence of latitudes (in degrees), :math:`\\phi_2`
    :param lon_2: A sequence of longitudes (in degrees), :math:`\\lambda_2`
    :returns: The bearing between two points
    """
    lon_1 = np.radians(lon_1)
    lat_1 = np.radians(lat_1)
    lon_2 = np.radians(lon_2)
    lat_2 = np.radians(lat_2)
    x = np.cos(lat_2) * np.sin(lon_2 - lon_1)
    y = np.cos(lat_1) * np.sin(lat_2) - np.sin(lat_1) * np.cos(lat_2) * np.cos(lon_2 - lon_1)
    return np.degrees(np.arctan2(x, y))


def reverse_bearing(lat_1: npt.NDArray[np.float64],
                    lon_1: npt.NDArray[np.float64],
                    distance: npt.NDArray[np.float64],
                    bearing: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """
    Compute the position of an object, given its initial position, the initial bearing, and the distance
    it will travel.
    See: https://www.igismap.com/formula-to-find-bearing-or-heading-angle-between-two-points-latitude-longitude/

    :param lat_1: A sequence of latitudes (in degrees)
    :param lon_1: A sequence of longitudes (in degrees)
    :param distance: The distance traveled for each ship
    :param bearing: The bearing (in degrees) for each ship

    :returns: The new position of the object
"""
    lon_1 = np.radians(lon_1)
    lat_1 = np.radians(lat_1)
    bearing = np.radians(bearing)
    ad = distance / EARTH_RADIUS
    lat_2 = np.arcsin(np.sin(lat_1) * np.cos(ad) + np.cos(lat_1) * np.sin(ad) * np.cos(bearing))
    lon_2 = lon_1 + np.arctan2(np.sin(bearing) * np.sin(ad) * np.cos(lat_1), np.cos(ad) - np.sin(lat_1) * np.sin(lat_2))
    return np.degrees(lat_2), np.degrees(lon_2)


@njit
def decdeg2dms(dd: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """
    Convert decimal degrees (dd) to sexagesimal degrees (degrees, minutes, seconds)

    Uses the formula from https://en.wikipedia.org/wiki/Decimal_degrees#Example:

    .. math::

        D &= \\mathrm{trunc}(dd, 0)\\\\
        M &= \\mathrm{trunc}(60 \\times \\vert dd - D\\vert, 0)\\\\
        S &= 3600 \\times \\vert dd - D \\vert - 60 \\times M\\\\

    :param dd: The decimal degrees
    :returns: The degrees :math:`D`, minutes :math:`M`, seconds of a decimal :math:`S`
    """
    D = np.trunc(dd)
    M = np.trunc(60 * np.abs(dd - D))
    S = 3600 * np.abs(dd - D) - 60 * M
    return D, M, S


def dms2decdeg(degrees: npt.NDArray[np.float64],
               minutes: npt.NDArray[np.float64],
               seconds: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """Convert sexagesimal degrees (degrees, minutes, seconds) to decimal degrees (dd).

    Uses the formula from https://en.wikipedia.org/wiki/Decimal_degrees#Example:

    .. math::

        dd = D + \\frac{M}{60} + \\frac{S}{3600}

    :param degrees: The degrees (:math:`D`)
    :param minutes: The minutes (:math:`M`)
    :param seconds: The seconds (:math:`S`)
    :returns: The decimal representation of the latitude/longitude.

    """
    # Get sign of any zero-degree input
    sign = np.ones_like(degrees)
    sign[np.signbit(degrees)] = -1
    return degrees + sign * (minutes / 60 + seconds / 3600)


def great_circle_distance(lat_1: npt.NDArray[np.float64],
                          lon_1: npt.NDArray[np.float64],
                          lat_2: npt.NDArray[np.float64],
                          lon_2: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """
    Computes the great circle distance between two points.

    Uses the :math:`(\\phi_1, \\lambda_1)`, :math:`(\\phi_2, \\lambda_2)` using the
    `Haversine formula <https://en.wikipedia.org/wiki/Haversine_formula>`_.

    .. math::

        d = 2 R \\arcsin \\left(\\sqrt{\\sin^2\\left(\\frac{\\phi_1 - \\phi_2}{2} \\right)
        + \\cos\\phi_1\\cos\\phi_2\\sin^2\\left(\\frac{\\lambda_1 - \\lambda_2}{2} \\right)
        }\\right)

    :param lat_1: A sequence of latitudes (in degrees), :math:`\\phi_1`
    :param lon_1: A sequence of longitudes (in degrees), :math:`\\lambda_1`
    :param lat_2: A sequence of latitudes (in degrees), :math:`\\phi_2`
    :param lon_2: A sequence of longitudes (in degrees), :math:`\\lambda_2`

    :returns: The great circle distance between sets of points """
    lon_1, lat_1, lon_2, lat_2 = map(np.radians, [lon_1, lat_1, lon_2, lat_2])
    return 2 * EARTH_RADIUS * np.arcsin(np.sqrt(np.power(np.sin((lat_1 - lat_2) / 2), 2)
                                                + np.cos(lat_1) * np.cos(lat_2) * np.power(np.sin((lon_1 - lon_2) / 2),
                                                                                           2)))


def chord_distance(d: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """
    Computes the chord length depending on the greater circle distance.

    This is the shortest distance inside the earth between two points.

    .. math::
        \\mathrm{crd} = 2 R \\sin \\left(\\theta/2 \\right)

    where :math:`\\theta=\\frac{d}{R}`, `d` being the great circle distance.

    :param d: The great circle distance
    :returns: The shortest distance between two points
    """
    return 2 * EARTH_RADIUS * np.sin(d / (2 * EARTH_RADIUS))


def angle_sat_c(c: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """
    Compute angle between sub-satellite location and vessel given the coord length `c`.

    Using the fact that the triangle made by the sub-longitude/latitude of the satellite
    and the straight line through the earth made by extending the line from the satellite through
    the earth's core can be used to determine the angle between the vessel, sub-long/lat of the satellite
    and the satellite.

    .. note::
        This angle is used in :func:`distance_sat_vessel` to determine the distance from the satellite to the vessel.

    :param c: The length of the chord going from the satellite sub-longitude/latitude to the vessel
    :returns: The angle between the vessel, satellite sub-longitude/latitude and the satellite.
    """
    return np.pi - np.arccos(c / (2 * EARTH_RADIUS))


def distance_sat_vessel(v: npt.NDArray[np.float64],
                        s: npt.NDArray[np.float64],
                        alpha: npt.NDArray[np.float64]) -> npt.NDArray[np.float64]:
    """
    Compute distance to satellite from vessel by using the theorem Al-Kashi.

    Given a satellite at Point S and (sub longitude, sub latitude) at Point U and a vessel
    at Point V, the distance from the satellite to the vessel :math:`u=\\vert S-V\\vert`
    is given by

    .. math::

        u^2= s^2 + v^2 - 2 s v \\cos\\alpha

    :param v: The altitude of the satellite over the earth :math:`\\vert U - S\\vert`
    :param s: The `s` is the chord length from the vessel to the satellite
              sub longitude/latitude  :math:`\\vert U - V\\vert`
    :param alpha: The angle :math:`\\alpha=\\angle SUV`
    :returns: The distance :math:`u` from the satellite to the vessel
    """
    return np.sqrt(np.power(s, 2) + np.power(v, 2) - 2 * s * v * np.cos(alpha))



from enum import IntEnum
from typing import List


class AISNavigationalStatus(IntEnum):
    """
    The AIS Navigational Status:

    :see https://help.marinetraffic.com/hc/en-us/articles/203990998-What-is-the-significance-of-the-AIS-Navigational-Status-Values-

    """  # noqa: E501
    UnderWayUsingEngine = 0
    AtAnchor = 1
    NotUnderCommand = 2
    RestrictedManeuverability = 3
    ConstrainedByHerDraught = 4
    Moored = 5
    Aground = 6
    EngagedInFishing = 7
    UnderWaySailing = 8
    _RESERVED_FOR_FUTURE__HAZARDOUS_GOODS_0 = 9
    _RESERVED_FOR_FUTURE__HAZARDOUS_GOODS_1 = 10
    Power_DrivenVesselTowingAstern = 11
    Power_DrivenVesselPushingAheadOrTowingAlongside = 12
    _RESERVED_FOR_FUTURE_USE = 13
    AIS_SART__MOB_AIS__EPIRB_AIS = 14
    Undefined = 15

    @classmethod
    def get_values(cls) -> List[int]:
        return [e.value for e in AISNavigationalStatus]



"""
Module containing the functionality for a Job Scheduler to run learning tasks
"""

from __future__ import annotations

import json
import select
import socket
import tempfile
from enum import Enum
from logging import INFO, Logger, getLogger
from pathlib import Path
from threading import Thread
from time import sleep
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

PREDICT_FILE_SOCKET = str(Path(tempfile.gettempdir()) / ".damast-predict")
_log: Logger = getLogger(__name__)


class Job:
    """
    Representation for a Prediction/Classification job
    """
    id: int

    #: The experiment directory which will be used to look for available models
    experiment_dir: str
    #: The model that shall be loaded
    model_name: str
    #: The (input) features
    features: List[str]
    #: The target features that will be predicted
    target: List[str]
    #: The length of the sequence
    sequence_length: int
    #: The path to the file containing the sequences to run the prediction on
    data_filename: str

    class Status(str, Enum):
        NOT_STARTED = "NOT STARTED"
        RUNNING = "RUNNING"
        STOPPED = "STOPPED"
        FINISHED = "FINISHED"

        FAILED = "FAILED"

    @classmethod
    def wait_for_status(cls, status_collector: Callable[[], Tuple[List[Job.Response], Job.Status]],
                        match_status: Optional[Job.Status] = None,
                        timeout_in_s: int = 10):
        for _ in range(timeout_in_s):
            collected_responses, current_status = status_collector()
            if match_status == current_status:
                return collected_responses

            sleep(1)

        raise TimeoutError(f"{cls.__name__}: no status: {match_status} encountered within {timeout_in_s} seconds")

    @classmethod
    def wait_for_responses(cls, status_collector: Callable[[], Tuple[List[Job.Response], Job.Status]],
                           predicate_responses: Callable[[List[Job.Response]], bool],
                           timeout_in_s: int = 10):
        for _ in range(timeout_in_s):
            collected_responses, _ = status_collector()
            if predicate_responses(collected_responses):
                return collected_responses

            sleep(1)

        raise TimeoutError(f"{cls.__name__}: no responses encountered that matched the condition")

    class Response:
        """
        A Response class which should be return when a job has finished.

        :param id: Identifier for the job that this response belong to
        :param timepoint: Related timepoint or index value for this sequence
        :param loss: The computed loss when comparing prediction with actual input
        :param actual_sequence: the actual sequence that shall be used for prediction
        :param predicted_sequence: the predicted features / sequence
        """
        #: id of the job that triggered this response
        id: int
        #: Timepoint / Index in the sequence
        timepoint: int
        #: The computed loss for this prediction
        loss: float
        #: The actual input sequence
        actual_sequence: List[Any]
        #: The predicted sequence or features
        predicted_sequence: List[Any]

        def __init__(self,
                     id: int,
                     timepoint: int,
                     loss: float,
                     actual_sequence: List[Any],
                     predicted_sequence: List[Any]):
            self.id = id
            self.timepoint = timepoint
            self.loss = loss
            self.actual_sequence = actual_sequence
            self.predicted_sequence = predicted_sequence

        def encode(self) -> bytes:
            """
            Encode this object into a transferable string
            :return: encoded string
            """
            return json.dumps(self.__dict__).encode()

        @classmethod
        def decode(cls, bytes: bytes) -> Job.Response:
            """
            Decode an encoded string / bytes into a :class:`Job.Response`
            :param bytes:
            :return: Instance of :class:`Job.Response`
            """
            values = json.loads(s=bytes.decode())
            return cls(**values)

        def __eq__(self, other):
            return self.__dict__ == other.__dict__

    def __init__(self,
                 id: int,
                 experiment_dir: Union[str, Path],
                 model_name: str,
                 features: List[str],
                 target: List[str],
                 sequence_length: int,
                 data_filename: Union[str, Path]):
        self.id = id
        self.experiment_dir = str(experiment_dir)
        self.model_name = model_name
        self.features = features
        self.target = target
        self.sequence_length = sequence_length
        self.data_filename = str(data_filename)

        # try to encode for validation purposes
        self.encode()

    def encode(self) -> bytes:
        return json.dumps(self.__dict__).encode()

    @classmethod
    def decode(cls, bytes: bytes) -> Job:
        values = json.loads(bytes.decode())
        return cls(**values)

    def __eq__(self, other) -> bool:
        return self.__dict__ == other.__dict__


class ControlCommand(str, Enum):
    BYE = "BYE"
    STOP = "STOP"


class ResponseCollector:
    job_id: int

    sock: socket.socket
    thread: Optional[Thread]

    responses: List[Job.Response]
    status: Job.Status

    def __init__(self,
                 job_id: int,
                 sock: socket.socket):
        self.job_id = job_id
        self.sock = sock
        self.thread = None

        self.status = Job.Status.NOT_STARTED
        self.responses = []

    def get_status(self) -> Tuple[List[Job.Response], Job.Status]:
        return self.responses, self.status

    def start(self):
        """
        Start the collection of results.
        :return:
        """
        self.thread = Thread(target=self.read_responses)
        self.thread.start()

    def read_responses(self):
        """
        Read all the responses from a given connection until the control command BYE is being received.
        """
        if self.sock is None:
            raise RuntimeError(f"{self.__class__.__name__}.read_responses: (job: {self.job_id}): no socket available")

        stop = False
        while not stop:
            self.status = Job.Status.RUNNING
            try:
                ready = select.select([self.sock], [], [], 0.01)
                if ready[0]:
                    # Get the length field first,
                    # then extract the data
                    msg = self.sock.recvmsg(4)

                    # Exit when encountering 'BYE' message
                    if msg[0] == ControlCommand.BYE.value.encode():
                        stop = True
                        break

                    msg_size = int.from_bytes(bytes=msg[0], byteorder="little")
                    received_bytes = self.sock.recvmsg(msg_size)[0]
                    if len(received_bytes) == 0:
                        raise ConnectionAbortedError(f"{self.__class__.__name__}.read_responses:"
                                                     f" connection was reset")

                    job_response = Job.Response.decode(bytes=received_bytes)
                    self.responses.append(job_response)
            except Exception as e:
                self.status = Job.Status.FAILED
                _log.warning(f"{self.__class__.__name__}.read_responses: processing failed. Job Id {self.job_id}"
                             f" -- {e}")
                raise

        if stop:
            self.status = Job.Status.STOPPED
        else:
            self.status = Job.Status.FINISHED


class JobScheduler:
    #: Main count of the current job id
    _job_id: int

    # region Job Mapping
    #: Mapping of job id to socket
    _sockets: Dict[int, socket.socket]
    #: Mapping of job id to job
    _jobs: Dict[int, Job]
    #: Mapping of job id to the collecting threads
    _collectors: Dict[int, ResponseCollector]

    # endregion

    def __init__(self):

        self._job_id = 0
        self._sockets: Dict[int, socket.socket] = {}

        self._jobs = {}
        self._collectors = {}

    def stop(self, job_id: int):
        """
        Send the STOP request to the Worker to stop the job related thread

        :param job_id: Job which shall be stopped
        """
        if job_id in self._sockets:
            sock = self._sockets[job_id]
            try:
                sock.sendmsg([ControlCommand.STOP.value.encode()])
            except BrokenPipeError:
                pass

    def start(self, job: Job) -> Job:
        """
        Trigger the execution of a job.

        This will send the job description to a worker which will start a job
        related thread.

        :param job: Job description.
        """
        soc = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)

        try:
            soc.connect(str(PREDICT_FILE_SOCKET))
        except ConnectionRefusedError as e:
            raise RuntimeError(f"{self.__class__.__name__}.start: no worker available") from e
        except FileNotFoundError as e:
            raise RuntimeError(f"{self.__class__.__name__}.start: no worker started") from e
        soc.setblocking(0)

        job.id = self._job_id
        self._job_id += 1

        self._sockets[job.id] = soc
        self._jobs[job.id] = job

        soc.sendmsg([job.encode()])

        # start collection thread at the same time
        self._collect(job.id)

    def get_status(self, job_id) -> Tuple[List[Job.Response], Job.Status]:
        """
        Get the current status of the job.

        :param job_id: id of the job for which the status shall be retrieved
        :return: List of response and status description of the collector
        """
        if job_id not in self._collectors:
            raise KeyError(f"{self.__class__.__name__}.get_status: no started job with id '{job_id}' found.")

        collector = self._collectors[job_id]
        return collector.get_status()

    def _collect(self, job_id: int):
        """
        Start the collection of the responses for a job

        :param job_id: Job id
        """
        collector = ResponseCollector(job_id, self._sockets[job_id])
        self._collectors[job_id] = collector
        collector.start()



"""
Module containing the worker functionality to perform a learning task
"""
import datetime as dt
import select
import socket
from logging import INFO, Logger, basicConfig, getLogger
from pathlib import Path
from threading import Event, Thread
from typing import Callable, Dict, List, Optional

import numpy as np
import pandas as pd
import polars as pl

from damast.data_handling.accessors import SequenceIterator
from damast.ml.experiments import Experiment
from damast.ml.scheduler import PREDICT_FILE_SOCKET, ControlCommand, Job

_log: Logger = getLogger(__name__)


class Worker:
    threads: Dict[int, Thread]
    stop_event: Event

    def __init__(self):
        self.threads = {}
        self.stop_event = Event()
        self.stop_event.clear()

    def predict(self,
                job_id: int,
                model_name: str,
                experiment_dir: Path,
                df: pd.DataFrame,
                features: List[str],
                target: List[str],
                sequence_length: int,
                connection: socket.socket,
                stop_event: Event,
                update_callback: Optional[Callable] = None
                ):
        models = Experiment.from_directory(experiment_dir)
        current_model = models[model_name]

        sta = SequenceIterator(df=df)

        # Generate a sequence of length + the intended forecasted values
        gen_predict = sta.to_keras_generator(features=features,
                                             target=target,
                                             sequence_length=sequence_length)

        loss = []
        timepoints = []
        timepoint = 0

        result = []

        try:
            _log.info(f"{self.__class__.__name__}({job_id}): run prediction")
            while True:
                if stop_event.is_set():
                    _log.info(f"{self.__class__.__name__}({job_id}): external stop request received")
                    break

                # Allow to stop the worker from outside
                ready = select.select([connection], [], [], 0.01)
                if ready[0]:
                    msg = connection.recvmsg(4)
                    # Get the length field
                    data = msg[0].decode()
                    if type(data) == str and data == ControlCommand.STOP.value:
                        break

                data = next(gen_predict)
                if data is None:
                    break

                X, y = data
                # Since the model requires 'batches' to be fed, provide one batch, which one step
                input_data = X[np.newaxis, :, :]
                predicted_sequence = current_model.predict(input_data,
                                                           steps=1,
                                                           verbose=0)

                actual_sequence = y
                computed_loss = float(current_model.loss(actual_sequence, predicted_sequence[0])[0])

                if update_callback is not None:
                    response = Job.Response(
                        id=job_id,
                        timepoint=timepoint,
                        loss=computed_loss,
                        predicted_sequence=predicted_sequence[0].tolist(),
                        actual_sequence=actual_sequence.tolist()
                    )
                    update_callback(response)

                loss.append(computed_loss)
                timepoints.append(timepoint)
                timepoint += 1
                result.append([timepoint, computed_loss, predicted_sequence[0], actual_sequence])
        except StopIteration:
            pass

        return result

    def listen_and_accept(self):
        das_socket_path = Path(PREDICT_FILE_SOCKET)
        das_socket_path.unlink(missing_ok=True)

        das_socket = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        das_socket.bind(PREDICT_FILE_SOCKET)

        while True:
            das_socket.listen(1)
            _log.info(f"{self.__class__.__name__}: waiting for connection on {PREDICT_FILE_SOCKET}...")
            try:
                connection, client_address = das_socket.accept()

                _log.info(f"{self.__class__.__name__}: receiving job information ...")
                msg = connection.recvmsg(2064)
                job = Job.decode(msg[0])

                _log.info(f"{self.__class__.__name__}: creating thread for {job}")
                t = Thread(target=self.execute, args=(connection, client_address, job, self.stop_event))
                t.name = f"{self.__class__.__name__}({job.id}"
                self.threads[job.id] = t
                t.start()
            except Exception:
                _log.info(f"{self.__class__.__name__}: exception encountered - requesting all threads to stop")
                self.stop_event.set()
                raise

    def execute(self, connection, client_address, job, stop_event: Event):
        _log.info(f"{self.__class__.__name__}.execute: starting job {job.id}")

        def update(response: Job.Response):
            encoded_response = response.encode()
            msg_size = len(encoded_response)

            _log.debug(f"{self.__class__.__name__}({job.id}): sendmsg: size: {msg_size}, response: {response}")
            connection.sendmsg([msg_size.to_bytes(4, 'little'), encoded_response])

        try:
            _log.info(f"{self.__class__.__name__}({job.id}): loading data")
            start = dt.datetime.utcnow()
            df = pl.scan_parquet(job.data_filename)
            _log.info(
                f"{self.__class__.__name__}({job.id}): loading data "
                f"[done after {(dt.datetime.utcnow() - start).total_seconds()} seconds]")

            self.predict(job_id=job.id,
                         model_name=job.model_name,
                         experiment_dir=job.experiment_dir,
                         features=job.features,
                         target=job.target,
                         sequence_length=job.sequence_length,
                         df=df,
                         connection=connection,
                         stop_event=stop_event,
                         update_callback=update)

            connection.sendmsg([ControlCommand.BYE.value.encode()])
        except BrokenPipeError as e:
            raise ConnectionError(f"{self.__class__.__name__}.run_prediction:"
                                  f" failed to communicate the results due to a connection error - {e}") from e
        finally:
            connection.close()


if __name__ == "__main__":
    basicConfig()
    _log.setLevel(INFO)

    w = Worker()
    w.listen_and_accept()
